const char checkpoint_daemon_lua[] =
"-- checkpoint_daemon.lua (internal file)\n"
"\n"
"local log = require 'log'\n"
"local fiber = require 'fiber'\n"
"local yaml = require 'yaml'\n"
"local errno = require 'errno'\n"
"local digest = require 'digest'\n"
"local pickle = require 'pickle'\n"
"\n"
"local PREFIX = 'checkpoint_daemon'\n"
"\n"
"local daemon = {\n"
"    checkpoint_interval = 0;\n"
"    fiber = nil;\n"
"    control = nil;\n"
"}\n"
"\n"
"local function snapshot()\n"
"    log.info(\"making snapshot...\")\n"
"    local s, e = pcall(function() box.snapshot() end)\n"
"    if not s then\n"
"        log.error(\"error while creating snapshot: %s\", e)\n"
"    end\n"
"end\n"
"\n"
"local function daemon_fiber(self)\n"
"    fiber.name(PREFIX, {truncate = true})\n"
"    log.info(\"started\")\n"
"\n"
"    --\n"
"    -- Add random offset to the initial period to avoid simultaneous\n"
"    -- snapshotting when multiple instances of tarantool are running\n"
"    -- on the same host.\n"
"    -- See https://github.com/tarantool/tarantool/issues/732\n"
"    --\n"
"    local random = pickle.unpack('i', digest.urandom(4))\n"
"    local offset = random % self.checkpoint_interval\n"
"    while true do\n"
"        local period = self.checkpoint_interval + offset\n"
"        -- maintain next_snapshot_time as a self member for testing purposes\n"
"        self.next_snapshot_time = fiber.time() + period\n"
"        log.info(\"scheduled the next snapshot at %s\",\n"
"                os.date(\"%c\", self.next_snapshot_time))\n"
"        local msg = self.control:get(period)\n"
"        if msg == 'shutdown' then\n"
"            break\n"
"        elseif msg == 'reload' then\n"
"            offset = random % self.checkpoint_interval\n"
"            log.info(\"reloaded\") -- continue\n"
"        elseif msg == nil and box.info.status == 'running' then\n"
"            snapshot()\n"
"            offset = 0\n"
"        end\n"
"    end\n"
"    self.next_snapshot_time = nil\n"
"    log.info(\"stopped\")\n"
"end\n"
"\n"
"local function reload(self)\n"
"    if self.checkpoint_interval > 0 then\n"
"        if self.control == nil then\n"
"            -- Start daemon\n"
"            self.control = fiber.channel()\n"
"            self.fiber = fiber.create(daemon_fiber, self)\n"
"            fiber.sleep(0)\n"
"        else\n"
"            -- Reload daemon\n"
"            self.control:put(\"reload\")\n"
"            --\n"
"            -- channel:put() doesn't block the writer if there\n"
"            -- is a ready reader. Give daemon fiber way so that\n"
"            -- it can execute before reload() returns to the caller.\n"
"            --\n"
"            fiber.sleep(0)\n"
"        end\n"
"    elseif self.control ~= nil then\n"
"        -- Shutdown daemon\n"
"        self.control:put(\"shutdown\")\n"
"        self.fiber = nil\n"
"        self.control = nil\n"
"        fiber.sleep(0) -- see comment above\n"
"    end\n"
"end\n"
"\n"
"setmetatable(daemon, {\n"
"    __index = {\n"
"        set_checkpoint_interval = function()\n"
"            daemon.checkpoint_interval = box.cfg.checkpoint_interval\n"
"            reload(daemon)\n"
"            return\n"
"        end,\n"
"    }\n"
"})\n"
"\n"
"if box.internal == nil then\n"
"    box.internal = { [PREFIX] = daemon }\n"
"else\n"
"    box.internal[PREFIX] = daemon\n"
"end\n"
""
;
