const char console_lua[] =
"-- console.lua -- internal file\n"
"\n"
"local internal = require('console')\n"
"local session_internal = require('box.internal.session')\n"
"local fiber = require('fiber')\n"
"local socket = require('socket')\n"
"local log = require('log')\n"
"local errno = require('errno')\n"
"local urilib = require('uri')\n"
"local yaml = require('yaml')\n"
"local net_box = require('net.box')\n"
"\n"
"local YAML_TERM = '\\n...\\n'\n"
"local PUSH_TAG_HANDLE = '!push!'\n"
"\n"
"local function format(status, ...)\n"
"    local err\n"
"    if status then\n"
"        -- serializer can raise an exception\n"
"        status, err = pcall(internal.format, ...)\n"
"        if status then\n"
"            return err\n"
"        else\n"
"            err = 'console: an exception occurred when formatting the output: '..\n"
"                tostring(err)\n"
"        end\n"
"    else\n"
"        err = ...\n"
"        if err == nil then\n"
"            err = box.NULL\n"
"        end\n"
"    end\n"
"    return internal.format({ error = err })\n"
"end\n"
"\n"
"--\n"
"-- Evaluate command on local instance\n"
"--\n"
"local function local_eval(self, line)\n"
"    if not line then\n"
"        return nil\n"
"    end\n"
"    --\n"
"    -- Attempt to append 'return ' before the chunk: if the chunk is\n"
"    -- an expression, this pushes results of the expression onto the\n"
"    -- stack. If the chunk is a statement, it won't compile. In that\n"
"    -- case try to run the original string.\n"
"    --\n"
"    local fun, errmsg = loadstring(\"return \"..line)\n"
"    if not fun then\n"
"        fun, errmsg = loadstring(line)\n"
"    end\n"
"    if not fun then\n"
"        return format(false, errmsg)\n"
"    end\n"
"    return format(pcall(fun))\n"
"end\n"
"\n"
"local function eval(line)\n"
"    return local_eval(nil, line)\n"
"end\n"
"\n"
"local text_connection_mt = {\n"
"    __index = {\n"
"        --\n"
"        -- Close underlying socket.\n"
"        --\n"
"        close = function(self)\n"
"            self._socket:close()\n"
"        end,\n"
"        --\n"
"        -- Write a text into a socket.\n"
"        -- @param test Text to send.\n"
"        -- @retval not nil Bytes sent.\n"
"        -- @retval     nil Error.\n"
"        --\n"
"        write = function(self, text)\n"
"            return self._socket:write(text)\n"
"        end,\n"
"        --\n"
"        -- Read a text from a socket until YAML terminator.\n"
"        -- @retval not nil Well formatted YAML.\n"
"        -- @retval     nil Error.\n"
"        --\n"
"        read = function(self)\n"
"            local ret = self._socket:read(YAML_TERM)\n"
"            if ret and ret ~= '' then\n"
"                return ret\n"
"            end\n"
"        end,\n"
"        --\n"
"        -- Write + Read.\n"
"        --\n"
"        eval = function(self, text)\n"
"            text = text..'$EOF$\\n'\n"
"            if not self:write(text) then\n"
"                error(self:set_error())\n"
"            end\n"
"            while true do\n"
"                local rc = self:read()\n"
"                if not rc then\n"
"                    break\n"
"                end\n"
"                local handle, prefix = yaml.decode(rc, {tag_only = true})\n"
"                if not handle then\n"
"                    -- Can not fail - tags are encoded with no\n"
"                    -- user participation and are correct always.\n"
"                    return rc\n"
"                end\n"
"                if handle == PUSH_TAG_HANDLE and self.print_f then\n"
"                    self.print_f(rc)\n"
"                end\n"
"            end\n"
"            return error(self:set_error())\n"
"        end,\n"
"        --\n"
"        -- Make the connection be in error state, set error\n"
"        -- message.\n"
"        -- @retval Error message.\n"
"        --\n"
"        set_error = function(self)\n"
"            self.state = 'error'\n"
"            self.error = self._socket:error()\n"
"            if not self.error then\n"
"                self.error = 'Peer closed'\n"
"            end\n"
"            return self.error\n"
"        end,\n"
"    }\n"
"}\n"
"\n"
"--\n"
"-- Wrap an existing socket with text Read-Write API inside a\n"
"-- netbox-like object.\n"
"-- @param connection Socket to wrap.\n"
"-- @param url Parsed destination URL.\n"
"-- @param print_f Function to print push messages.\n"
"--\n"
"-- @retval nil, err Error, and err contains an error message.\n"
"-- @retval  not nil Netbox-like object.\n"
"--\n"
"local function wrap_text_socket(connection, url, print_f)\n"
"    local conn = setmetatable({\n"
"        _socket = connection,\n"
"        state = 'active',\n"
"        host = url.host or 'localhost',\n"
"        port = url.service,\n"
"        print_f = print_f,\n"
"    }, text_connection_mt)\n"
"    if not conn:write('require(\"console\").delimiter(\"$EOF$\")\\n') or\n"
"       not conn:read() then\n"
"        conn:set_error()\n"
"    end\n"
"    return conn\n"
"end\n"
"\n"
"--\n"
"-- Evaluate command on remote instance\n"
"--\n"
"local function remote_eval(self, line)\n"
"    if line and self.remote.state == 'active' then\n"
"        local ok, res = pcall(self.remote.eval, self.remote, line)\n"
"        if self.remote.state == 'active' then\n"
"            return ok and res or format(false, res)\n"
"        end\n"
"    end\n"
"    local err = self.remote.error\n"
"    self.remote:close()\n"
"    self.remote = nil\n"
"    self.eval = nil\n"
"    self.prompt = nil\n"
"    self.completion = nil\n"
"    pcall(self.on_client_disconnect, self)\n"
"    return (err and format(false, err)) or ''\n"
"end\n"
"\n"
"--\n"
"-- Read command from stdin\n"
"--\n"
"local function local_read(self)\n"
"    local buf = \"\"\n"
"    local prompt = self.prompt\n"
"    while true do\n"
"        local delim = self.delimiter\n"
"        local line = internal.readline({\n"
"            prompt = prompt.. \"> \",\n"
"            completion = self.ac and self.completion or nil\n"
"        })\n"
"        if not line then\n"
"            return nil\n"
"        end\n"
"        buf = buf..line\n"
"        if delim == \"\" then\n"
"            -- stop once a complete Lua statement is entered\n"
"            local fn, err = loadstring(buf)\n"
"            if fn ~= nil or not string.find(err, \" near '<eof>'$\") then\n"
"                -- valid Lua code or a syntax error not due to\n"
"                -- an incomplete input\n"
"                break\n"
"            end\n"
"            if loadstring('return '..buf) ~= nil then\n"
"                -- certain obscure inputs like '(42\\n)' yield the\n"
"                -- same error as incomplete statement\n"
"                break\n"
"            end\n"
"        elseif #buf >= #delim and buf:sub(#buf - #delim + 1) == delim then\n"
"            buf = buf:sub(0, #buf - #delim)\n"
"            break\n"
"        end\n"
"        buf = buf..\"\\n\"\n"
"        prompt = string.rep(' ', #self.prompt)\n"
"    end\n"
"    internal.add_history(buf)\n"
"    if self.history_file then\n"
"        internal.save_history(self.history_file)\n"
"    end\n"
"    return buf\n"
"end\n"
"\n"
"--\n"
"-- Print result to stdout\n"
"--\n"
"local function local_print(self, output)\n"
"    if output == nil then\n"
"        self.running = nil\n"
"        return\n"
"    end\n"
"    print(output)\n"
"end\n"
"\n"
"--\n"
"-- Read command from connected client console.listen()\n"
"--\n"
"local function client_read(self)\n"
"    local delim = self.delimiter..\"\\n\"\n"
"    local buf = self.client:read(delim)\n"
"    if buf == nil then\n"
"        return nil\n"
"    elseif buf == \"\" then\n"
"        return nil -- EOF\n"
"    elseif buf == \"~.\\n\" then\n"
"        -- Escape sequence to close current connection (like SSH)\n"
"        return nil\n"
"    end\n"
"    -- remove trailing delimiter\n"
"    return buf:sub(1, -#delim-1)\n"
"end\n"
"\n"
"--\n"
"-- Print result to connected client from console.listen()\n"
"--\n"
"local function client_print(self, output)\n"
"    if not self.client then\n"
"        return\n"
"    elseif not output then\n"
"        -- disconnect peer\n"
"        self.client = nil -- socket will be closed by tcp_server() function\n"
"        self.running = nil\n"
"        return\n"
"    end\n"
"    self.client:write(output)\n"
"end\n"
"\n"
"--\n"
"-- REPL state\n"
"--\n"
"local repl_mt = {\n"
"    __index = {\n"
"        running = false;\n"
"        delimiter = \"\";\n"
"        prompt = \"tarantool\";\n"
"        read = local_read;\n"
"        eval = local_eval;\n"
"        print = local_print;\n"
"        completion = internal.completion_handler;\n"
"        ac = true;\n"
"    };\n"
"}\n"
"\n"
"--\n"
"-- REPL = read-eval-print-loop\n"
"--\n"
"local function repl(self)\n"
"    fiber.self().storage.console = self\n"
"    if type(self.on_start) == 'function' then\n"
"        self:on_start()\n"
"    end\n"
"\n"
"    while self.running do\n"
"        local command = self:read()\n"
"        local output = self:eval(command)\n"
"        self:print(output)\n"
"    end\n"
"    fiber.self().storage.console = nil\n"
"end\n"
"\n"
"local function on_start(foo)\n"
"    if foo == nil or type(foo) == 'function' then\n"
"        repl_mt.__index.on_start = foo\n"
"        return\n"
"    end\n"
"    error('Wrong type of on_start hook: ' .. type(foo))\n"
"end\n"
"\n"
"local function on_client_disconnect(foo)\n"
"    if foo == nil or type(foo) == 'function' then\n"
"        repl_mt.__index.on_client_disconnect = foo\n"
"        return\n"
"    end\n"
"    error('Wrong type of on_client_disconnect hook: ' .. type(foo))\n"
"end\n"
"\n"
"--\n"
"-- Set delimiter\n"
"--\n"
"local function delimiter(delim)\n"
"    local self = fiber.self().storage.console\n"
"    if self == nil then\n"
"        error(\"console.delimiter(): need existing console\")\n"
"    end\n"
"    if delim == nil then\n"
"        return self.delimiter\n"
"    elseif type(delim) == 'string' then\n"
"        self.delimiter = delim\n"
"    else\n"
"        error('invalid delimiter')\n"
"    end\n"
"end\n"
"\n"
"--\n"
"--\n"
"--\n"
"local function ac(yes_no)\n"
"    local self = fiber.self().storage.console\n"
"    if self == nil then\n"
"        error(\"console.ac(): need existing console\")\n"
"    end\n"
"    self.ac = not not yes_no\n"
"end\n"
"\n"
"--\n"
"-- Start REPL on stdin\n"
"--\n"
"local started = false\n"
"local function start()\n"
"    if started then\n"
"        error(\"console is already started\")\n"
"    end\n"
"    started = true\n"
"    local self = setmetatable({ running = true }, repl_mt)\n"
"    local home_dir = os.getenv('HOME')\n"
"    if home_dir then\n"
"        self.history_file = home_dir .. '/.tarantool_history'\n"
"        internal.load_history(self.history_file)\n"
"    end\n"
"    session_internal.create(1, \"repl\") -- stdin fileno\n"
"    repl(self)\n"
"    started = false\n"
"end\n"
"\n"
"--\n"
"-- Connect to remove instance\n"
"--\n"
"local function connect(uri, opts)\n"
"    opts = opts or {}\n"
"\n"
"    local self = fiber.self().storage.console\n"
"    if self == nil then\n"
"        error(\"console.connect() need existing console\")\n"
"    end\n"
"\n"
"    local u\n"
"    if uri then\n"
"        u = urilib.parse(tostring(uri))\n"
"    end\n"
"    if u == nil or u.service == nil then\n"
"        error('Usage: console.connect(\"[login:password@][host:]port\")')\n"
"    end\n"
"\n"
"    local connection, greeting =\n"
"        net_box.establish_connection(u.host, u.service, opts.timeout)\n"
"    if not connection then\n"
"        log.verbose(greeting)\n"
"        box.error(box.error.NO_CONNECTION)\n"
"    end\n"
"    local remote\n"
"    if greeting.protocol == 'Lua console' then\n"
"        remote = wrap_text_socket(connection, u,\n"
"                                  function(msg) self:print(msg) end)\n"
"    else\n"
"        opts = {\n"
"            connect_timeout = opts.timeout,\n"
"            user = u.login,\n"
"            password = u.password,\n"
"        }\n"
"        remote = net_box.wrap(connection, greeting, u.host, u.service, opts)\n"
"        if not remote.host then\n"
"            remote.host = 'localhost'\n"
"        end\n"
"        local old_eval = remote.eval\n"
"        remote.eval = function(con, line)\n"
"            return old_eval(con, 'return require(\"console\").eval(...)', {line})\n"
"        end\n"
"    end\n"
"\n"
"    -- check connection && permissions\n"
"    local ok, res = pcall(remote.eval, remote, 'return true')\n"
"    if not ok then\n"
"        remote:close()\n"
"        pcall(self.on_client_disconnect, self)\n"
"        error(res)\n"
"    end\n"
"\n"
"    -- override methods\n"
"    self.remote = remote\n"
"    self.eval = remote_eval\n"
"    self.prompt = string.format(\"%s:%s\", self.remote.host, self.remote.port)\n"
"    self.completion = function (str, pos1, pos2)\n"
"        local c = string.format(\n"
"            'return require(\"console\").completion_handler(%q, %d, %d)',\n"
"            str, pos1, pos2)\n"
"        return yaml.decode(remote:eval(c))[1]\n"
"    end\n"
"    log.info(\"connected to %s:%s\", self.remote.host, self.remote.port)\n"
"    return true\n"
"end\n"
"\n"
"local function client_handler(client, peer)\n"
"    session_internal.create(client:fd(), \"console\")\n"
"    session_internal.run_on_connect()\n"
"    session_internal.run_on_auth(box.session.user(), true)\n"
"    local state = setmetatable({\n"
"        running = true;\n"
"        read = client_read;\n"
"        print = client_print;\n"
"        client = client;\n"
"    }, repl_mt)\n"
"    local version = _TARANTOOL:match(\"([^-]+)-\")\n"
"    state:print(string.format(\"%-63s\\n%-63s\\n\",\n"
"        \"Tarantool \".. version..\" (Lua console)\",\n"
"        \"type 'help' for interactive help\"))\n"
"    repl(state)\n"
"    session_internal.run_on_disconnect()\n"
"end\n"
"\n"
"--\n"
"-- Start admin console\n"
"--\n"
"local function listen(uri)\n"
"    local host, port\n"
"    if uri == nil then\n"
"        host = 'unix/'\n"
"        port = '/tmp/tarantool-console.sock'\n"
"    else\n"
"        local u = urilib.parse(tostring(uri))\n"
"        if u == nil or u.service == nil then\n"
"            error('Usage: console.listen(\"[host:]port\")')\n"
"        end\n"
"        host = u.host\n"
"        port = u.service or 3313\n"
"    end\n"
"    local s, addr = socket.tcp_server(host, port, { handler = client_handler,\n"
"        name = 'console'})\n"
"    if not s then\n"
"        error(string.format('failed to create server %s:%s: %s',\n"
"            host, port, errno.strerror()))\n"
"    end\n"
"    return s\n"
"end\n"
"\n"
"package.loaded['console'] = {\n"
"    start = start;\n"
"    eval = eval;\n"
"    delimiter = delimiter;\n"
"    ac = ac;\n"
"    connect = connect;\n"
"    listen = listen;\n"
"    on_start = on_start;\n"
"    on_client_disconnect = on_client_disconnect;\n"
"    completion_handler = internal.completion_handler;\n"
"}\n"
""
;
