const char feedback_daemon_lua[] =
"-- feedback_daemon.lua (internal file)\n"
"--\n"
"local log   = require('log')\n"
"local json  = require('json')\n"
"local fiber = require('fiber')\n"
"local http  = require('http.client')\n"
"\n"
"local PREFIX = \"feedback_daemon\"\n"
"\n"
"local daemon = {\n"
"    enabled  = false,\n"
"    interval = 0,\n"
"    host     = nil,\n"
"    fiber    = nil,\n"
"    control  = nil,\n"
"    guard    = nil,\n"
"    shutdown = nil\n"
"}\n"
"\n"
"local function get_fiber_id(f)\n"
"    local fid = 0\n"
"    if f ~= nil and f:status() ~= \"dead\" then\n"
"        fid = f:id()\n"
"    end\n"
"    return fid\n"
"end\n"
"\n"
"local function fill_in_feedback(feedback)\n"
"    if box.info.status ~= \"running\" then\n"
"        return nil, \"not running\"\n"
"    end\n"
"    feedback.tarantool_version = box.info.version\n"
"    feedback.server_id         = box.info.uuid\n"
"    feedback.cluster_id        = box.info.cluster.uuid\n"
"    return feedback\n"
"end\n"
"\n"
"local function feedback_loop(self)\n"
"    fiber.name(PREFIX, { truncate = true })\n"
"\n"
"    while true do\n"
"        local feedback = self:generate_feedback()\n"
"        local msg = self.control:get(self.interval)\n"
"        -- if msg == \"send\" then we simply send feedback\n"
"        if msg == \"stop\" then\n"
"            break\n"
"        elseif feedback ~= nil then\n"
"            pcall(http.post, self.host, json.encode(feedback), {timeout=1})\n"
"        end\n"
"    end\n"
"    self.shutdown:put(\"stopped\")\n"
"end\n"
"\n"
"local function guard_loop(self)\n"
"    fiber.name(string.format(\"guard of %s\", PREFIX), {truncate=true})\n"
"\n"
"    while true do\n"
"\n"
"        if get_fiber_id(self.fiber) == 0 then\n"
"            self.fiber = fiber.create(feedback_loop, self)\n"
"            log.verbose(\"%s restarted\", PREFIX)\n"
"        end\n"
"        local st, err = pcall(fiber.sleep, self.interval)\n"
"        if not st then\n"
"            -- fiber was cancelled\n"
"            break\n"
"        end\n"
"    end\n"
"    self.shutdown:put(\"stopped\")\n"
"end\n"
"\n"
"-- these functions are used for test purposes only\n"
"local function start(self)\n"
"    self:stop()\n"
"    if self.enabled then\n"
"        self.control = fiber.channel()\n"
"        self.shutdown = fiber.channel()\n"
"        self.guard = fiber.create(guard_loop, self)\n"
"    end\n"
"    log.verbose(\"%s started\", PREFIX)\n"
"end\n"
"\n"
"local function stop(self)\n"
"    if (get_fiber_id(self.guard) ~= 0) then\n"
"        self.guard:cancel()\n"
"        self.shutdown:get()\n"
"    end\n"
"    if (get_fiber_id(self.fiber) ~= 0) then\n"
"        self.control:put(\"stop\")\n"
"        self.shutdown:get()\n"
"    end\n"
"    self.guard = nil\n"
"    self.fiber = nil\n"
"    self.control = nil\n"
"    self.shutdown = nil\n"
"    log.verbose(\"%s stopped\", PREFIX)\n"
"end\n"
"\n"
"local function reload(self)\n"
"    self:stop()\n"
"    self:start()\n"
"end\n"
"\n"
"setmetatable(daemon, {\n"
"    __index = {\n"
"        set_feedback_params = function()\n"
"            daemon.enabled  = box.cfg.feedback_enabled\n"
"            daemon.host     = box.cfg.feedback_host\n"
"            daemon.interval = box.cfg.feedback_interval\n"
"            reload(daemon)\n"
"            return\n"
"        end,\n"
"        -- this function is used in saving feedback in file\n"
"        generate_feedback = function()\n"
"            return fill_in_feedback({ feedback_type = \"version\", feedback_version = 1 })\n"
"        end,\n"
"        start = function()\n"
"            start(daemon)\n"
"        end,\n"
"        stop = function()\n"
"            stop(daemon)\n"
"        end,\n"
"        reload = function()\n"
"            reload(daemon)\n"
"        end,\n"
"        send_test = function()\n"
"            if daemon.control ~= nil then\n"
"                daemon.control:put(\"send\")\n"
"            end\n"
"        end\n"
"    }\n"
"})\n"
"\n"
"if box.internal == nil then\n"
"    box.internal = { [PREFIX] = daemon }\n"
"else\n"
"    box.internal[PREFIX] = daemon\n"
"end\n"
""
;
