const char upgrade_lua[] =
"local log = require('log')\n"
"local bit = require('bit')\n"
"local json = require('json')\n"
"\n"
"-- Guest user id - the default user\n"
"local GUEST = 0\n"
"-- Super User ID\n"
"local ADMIN = 1\n"
"-- role 'PUBLIC' is special, it's automatically granted to every user\n"
"local PUBLIC = 2\n"
"-- role 'REPLICATION'\n"
"local REPLICATION = 3\n"
"-- role 'SUPER'\n"
"-- choose a fancy id to not clash with any existing role or\n"
"-- user during upgrade\n"
"local SUPER = 31\n"
"\n"
"--------------------------------------------------------------------------------\n"
"-- Utils\n"
"--------------------------------------------------------------------------------\n"
"\n"
"local function setmap(tab)\n"
"    return setmetatable(tab, { __serialize = 'map' })\n"
"end\n"
"\n"
"local function ismap(tab)\n"
"    if type(tab) ~= 'table' then\n"
"        return false\n"
"    end\n"
"    local mt = getmetatable(tab)\n"
"    return mt and (mt.__serialize == 'map' or mt.__serialize == 'mapping')\n"
"end\n"
"\n"
"local mkversion = {}\n"
"mkversion.__index = mkversion\n"
"setmetatable(mkversion, {__call = function(c, ...) return c.new(...) end})\n"
"\n"
"function mkversion.new(major, minor, patch)\n"
"    local self = setmetatable({}, mkversion)\n"
"    self.major = major\n"
"    self.minor = minor\n"
"    self.patch = patch\n"
"    self.id = bit.bor(bit.lshift(bit.bor(bit.lshift(major, 8), minor), 8), patch)\n"
"    return self\n"
"end\n"
"\n"
"function mkversion.__tostring(self)\n"
"    return string.format('%s.%s.%s', self.major, self.minor, self.patch)\n"
"end\n"
"\n"
"function mkversion.__eq(lhs, rhs)\n"
"    return lhs.id == rhs.id\n"
"end\n"
"\n"
"function mkversion.__lt(lhs, rhs)\n"
"    return lhs.id < rhs.id\n"
"end\n"
"\n"
"-- space:truncate() doesn't work with disabled triggers on __index\n"
"local function truncate(space)\n"
"    local pk = space.index[0]\n"
"    while pk:len() > 0 do\n"
"        local state, t\n"
"        for state, t in pk:pairs() do\n"
"            local key = {}\n"
"            for _k2, parts in ipairs(pk.parts) do\n"
"                table.insert(key, t[parts.fieldno])\n"
"            end\n"
"            space:delete(key)\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"local function set_system_triggers(val)\n"
"    box.space._space:run_triggers(val)\n"
"    box.space._index:run_triggers(val)\n"
"    box.space._user:run_triggers(val)\n"
"    box.space._func:run_triggers(val)\n"
"    box.space._priv:run_triggers(val)\n"
"    box.space._schema:run_triggers(val)\n"
"    box.space._cluster:run_triggers(val)\n"
"end\n"
"\n"
"--------------------------------------------------------------------------------\n"
"-- Bootstrap\n"
"--------------------------------------------------------------------------------\n"
"\n"
"local function erase()\n"
"    truncate(box.space._space)\n"
"    truncate(box.space._index)\n"
"    truncate(box.space._user)\n"
"    truncate(box.space._func)\n"
"    truncate(box.space._priv)\n"
"    truncate(box.space._sequence_data)\n"
"    truncate(box.space._sequence)\n"
"    truncate(box.space._truncate)\n"
"    truncate(box.space._collation)\n"
"    truncate(box.space._schema)\n"
"    truncate(box.space._cluster)\n"
"end\n"
"\n"
"local function initial()\n"
"    -- stick to the following convention:\n"
"    -- prefer user id (owner id) in field #1\n"
"    -- prefer object name in field #2\n"
"    -- index on owner id is index #1\n"
"    -- index on object name is index #2\n"
"    --\n"
"\n"
"    local _schema = box.space[box.schema.SCHEMA_ID]\n"
"    local _space = box.space[box.schema.SPACE_ID]\n"
"    local _index = box.space[box.schema.INDEX_ID]\n"
"    local _func = box.space[box.schema.FUNC_ID]\n"
"    local _user = box.space[box.schema.USER_ID]\n"
"    local _priv = box.space[box.schema.PRIV_ID]\n"
"    local _cluster = box.space[box.schema.CLUSTER_ID]\n"
"    local MAP = setmap({})\n"
"\n"
"    --\n"
"    -- _schema\n"
"    --\n"
"    log.info(\"create space _schema\")\n"
"    _space:insert{_schema.id, ADMIN, '_schema', 'memtx', 0, MAP, {}}\n"
"    log.info(\"create index primary on _schema\")\n"
"    _index:insert{_schema.id, 0, 'primary', 'tree', { unique = true }, {{0, 'str'}}}\n"
"\n"
"    --\n"
"    -- _space\n"
"    --\n"
"    log.info(\"create space _space\")\n"
"    _space:insert{_space.id, ADMIN, '_space', 'memtx', 0, MAP, {}}\n"
"    -- space name is unique\n"
"    log.info(\"create index primary on _space\")\n"
"    _index:insert{_space.id, 0, 'primary', 'tree', { unique = true }, {{0, 'num'}}}\n"
"    log.info(\"create index owner on _space\")\n"
"    _index:insert{_space.id, 1, 'owner', 'tree', {unique = false }, {{1, 'num'}}}\n"
"    log.info(\"create index index name on _space\")\n"
"    _index:insert{_space.id, 2, 'name', 'tree', { unique = true }, {{2, 'str'}}}\n"
"\n"
"    --\n"
"    -- _index\n"
"    --\n"
"    log.info(\"create space _index\")\n"
"    _space:insert{_index.id, ADMIN, '_index', 'memtx', 0, MAP, {}}\n"
"    -- index name is unique within a space\n"
"    log.info(\"create index primary on _index\")\n"
"    _index:insert{_index.id, 0, 'primary', 'tree', {unique = true}, {{0, 'num'}, {1, 'num'}}}\n"
"    log.info(\"create index name on _index\")\n"
"    _index:insert{_index.id, 2, 'name', 'tree', {unique = true}, {{0, 'num'}, {2, 'str'}}}\n"
"\n"
"    --\n"
"    -- _func\n"
"    --\n"
"    log.info(\"create space _func\")\n"
"    _space:insert{_func.id, ADMIN, '_func', 'memtx', 0, MAP, {}}\n"
"    -- function name and id are unique\n"
"    log.info(\"create index _func:primary\")\n"
"    _index:insert{_func.id, 0, 'primary', 'tree', {unique = true}, {{0, 'num'}}}\n"
"    log.info(\"create index _func:owner\")\n"
"    _index:insert{_func.id, 1, 'owner', 'tree', {unique = false}, {{1, 'num'}}}\n"
"    log.info(\"create index _func:name\")\n"
"    _index:insert{_func.id, 2, 'name', 'tree', {unique = true}, {{2, 'str'}}}\n"
"\n"
"    --\n"
"    -- _user\n"
"    --\n"
"    log.info(\"create space _user\")\n"
"    _space:insert{_user.id, ADMIN, '_user', 'memtx', 0, MAP, {}}\n"
"    -- user name and id are unique\n"
"    log.info(\"create index _func:primary\")\n"
"    _index:insert{_user.id, 0, 'primary', 'tree', {unique = true}, {{0, 'num'}}}\n"
"    log.info(\"create index _func:owner\")\n"
"    _index:insert{_user.id, 1, 'owner', 'tree', {unique = false}, {{1, 'num'}}}\n"
"    log.info(\"create index _func:name\")\n"
"    _index:insert{_user.id, 2, 'name', 'tree', {unique = true}, {{2, 'str'}}}\n"
"\n"
"    --\n"
"    -- _priv\n"
"    --\n"
"    log.info(\"create space _priv\")\n"
"    _space:insert{_priv.id, ADMIN, '_priv', 'memtx', 0, MAP, {}}\n"
"    --\n"
"    -- space schema is: grantor id, user id, object_type, object_id, privilege\n"
"    -- primary key: user id, object type, object id\n"
"    log.info(\"create index primary on _priv\")\n"
"    _index:insert{_priv.id, 0, 'primary', 'tree', {unique = true}, {{1, 'num'}, {2, 'str'}, {3, 'num'}}}\n"
"    -- owner index  - to quickly find all privileges granted by a user\n"
"    log.info(\"create index owner on _priv\")\n"
"    _index:insert{_priv.id, 1, 'owner', 'tree', {unique = false}, {{0, 'num'}}}\n"
"    -- object index - to quickly find all grants on a given object\n"
"    log.info(\"create index object on _priv\")\n"
"    _index:insert{_priv.id, 2, 'object', 'tree', {unique = false}, {{2, 'str'}, {3, 'num'}}}\n"
"\n"
"    --\n"
"    -- _cluster\n"
"    --\n"
"    log.info(\"create space _cluster\")\n"
"    _space:insert{_cluster.id, ADMIN, '_cluster', 'memtx', 0, MAP, {}}\n"
"    -- primary key: node id\n"
"    log.info(\"create index primary on _cluster\")\n"
"    _index:insert{_cluster.id, 0, 'primary', 'tree', {unique = true}, {{0, 'num'}}}\n"
"    -- node uuid key: node uuid\n"
"    log.info(\"create index uuid on _cluster\")\n"
"    _index:insert{_cluster.id, 1, 'uuid', 'tree', {unique = true}, {{1, 'str'}}}\n"
"\n"
"    --\n"
"    -- Pre-create user and grants\n"
"    log.info(\"create user guest\")\n"
"    _user:insert{GUEST, ADMIN, 'guest', 'user'}\n"
"    log.info(\"create user admin\")\n"
"    _user:insert{ADMIN, ADMIN, 'admin', 'user'}\n"
"    log.info(\"create role public\")\n"
"    _user:insert{PUBLIC, ADMIN, 'public', 'role'}\n"
"    log.info(\"grant read,write,execute on universe to admin\")\n"
"    _priv:insert{ADMIN, ADMIN, 'universe', 0, box.priv.R + box.priv.W + box.priv.X}\n"
"\n"
"    -- grant role 'public' to 'guest'\n"
"    log.info(\"grant role public to guest\")\n"
"    _priv:insert{ADMIN, GUEST, 'role', PUBLIC, box.priv.X}\n"
"\n"
"    log.info(\"set max_id to box.schema.SYSTEM_ID_MAX\")\n"
"    _schema:insert{'max_id', box.schema.SYSTEM_ID_MAX}\n"
"\n"
"    log.info(\"set schema version to 1.6.0\")\n"
"    _schema:insert({'version', 1, 6, 0})\n"
"end\n"
"\n"
"--------------------------------------------------------------------------------\n"
"-- Tarantool 1.6.8\n"
"--------------------------------------------------------------------------------\n"
"\n"
"local function upgrade_index_options_to_1_6_8()\n"
"    local indexes = {}\n"
"    for _, def in box.space._index:pairs() do\n"
"        if type(def[5]) == 'number' then\n"
"            -- Tarantool < 1.6.5 format\n"
"            local part_count = def[6]\n"
"            local new_def = def:update({{'#', 7, 2 * part_count}}):totable()\n"
"            new_def[5] = setmap({})\n"
"            new_def[5].unique = def[5] ~= 0\n"
"            new_def[6] = {}\n"
"            for i=1,part_count,1 do\n"
"                local field_id = def[7 + (i - 1) * 2]\n"
"                local field_type = def[7 + (i - 1) * 2 + 1]\n"
"                table.insert(new_def[6], { field_id, field_type })\n"
"            end\n"
"            table.insert(indexes, new_def)\n"
"        elseif not ismap(def[5]) then\n"
"            log.error(\"unexpected index options: %s\", json.encode(def[5]))\n"
"        end\n"
"    end\n"
"    for _, new_def in ipairs(indexes) do\n"
"        log.info(\"alter index %s on %s set options to %s, parts to %s\",\n"
"                 new_def[3], box.space[new_def[1]].name,\n"
"                 json.encode(new_def[5]), json.encode(new_def[6]))\n"
"        box.space._index:replace(new_def)\n"
"    end\n"
"end\n"
"\n"
"local function upgrade_space_options_to_1_6_8()\n"
"    local spaces = {}\n"
"    for _, def in box.space._space:pairs() do\n"
"        local new_def = def:totable()\n"
"        new_def[6] = setmap({})\n"
"        if def[6] == nil or def[6] == \"\" then\n"
"            -- Tarantool < 1.6.8 format\n"
"            table.insert(spaces, new_def)\n"
"        elseif def[6] == 'temporary' then\n"
"            -- Tarantool < 1.6.8 format\n"
"            new_def[6].temporary = true\n"
"            table.insert(spaces, new_def)\n"
"        elseif not ismap(def[6]) then\n"
"            log.error(\"unexpected space options: %s\", json.encode(def[6]))\n"
"        end\n"
"    end\n"
"    for _, new_def in ipairs(spaces) do\n"
"        log.info(\"alter space %s set options to %s\", new_def[3],\n"
"                 json.encode(new_def[6]))\n"
"        box.space._space:update(new_def[1], {{'=', 6, new_def[6]}})\n"
"    end\n"
"end\n"
"\n"
"local function upgrade_space_format_to_1_6_8()\n"
"    local space_def = box.space._space:get(box.space._schema.id)\n"
"    if space_def[7] == nil or next(space_def[7]) == nil then\n"
"        local format = {}\n"
"        format[1] = {type='str', name='key'}\n"
"        log.info(\"alter space _schema set format to %s\", json.encode(format))\n"
"        box.space._space:update(box.space._schema.id, {{'=', 7, format}})\n"
"    end\n"
"\n"
"    local space_def = box.space._space:get(box.space._space.id)\n"
"    if space_def[7] == nil or next(space_def[7]) == nil then\n"
"        local format = {}\n"
"        format[1] = {name='id', type='num'}\n"
"        format[2] = {name='owner', type='num'}\n"
"        format[3] = {name='name', type='str'}\n"
"        format[4] = {name='engine', type='str'}\n"
"        format[5] = {name='field_count', type='num'}\n"
"        format[6] = {name='flags', type='str'}\n"
"        format[7] = {name='format', type='*'}\n"
"        log.info(\"alter space _space set format\")\n"
"        box.space._space:format(format)\n"
"    end\n"
"\n"
"    local space_def = box.space._space:get(box.space._index.id)\n"
"    if space_def[7] == nil or next(space_def[7]) == nil or\n"
"       space_def[7][5].name == 'unique' then\n"
"        local format = {}\n"
"        format[1] = {name = 'id', type = 'num'}\n"
"        format[2] = {name = 'iid', type = 'num'}\n"
"        format[3] = {name = 'name', type = 'str'}\n"
"        format[4] = {name = 'type', type = 'str'}\n"
"        format[5] = {name = 'opts', type = 'array'}\n"
"        format[6] = {name = 'parts', type = 'array'}\n"
"        log.info(\"alter space _index set format\")\n"
"        box.space._index:format(format)\n"
"    end\n"
"\n"
"    local space_def = box.space._space:get(box.space._func.id)\n"
"    if space_def[7] == nil or next(space_def[7]) == nil then\n"
"        local format = {}\n"
"        format[1] = {name='id', type='num'}\n"
"        format[2] = {name='owner', type='num'}\n"
"        format[3] = {name='name', type='str'}\n"
"        format[4] = {name='setuid', type='num'}\n"
"        log.info(\"alter space _func set format\")\n"
"        box.space._func:format(format)\n"
"    end\n"
"\n"
"    local space_def = box.space._space:get(box.space._user.id)\n"
"    if space_def[7] == nil or next(space_def[7]) == nil then\n"
"        local format = {}\n"
"        format[1] = {name='id', type='num'}\n"
"        format[2] = {name='owner', type='num'}\n"
"        format[3] = {name='name', type='str'}\n"
"        format[4] = {name='type', type='str'}\n"
"        format[5] = {name='auth', type='*'}\n"
"        log.info(\"alter space _user set format\")\n"
"        box.space._user:format(format)\n"
"    end\n"
"\n"
"    local space_def = box.space._space:get(box.space._priv.id)\n"
"    if space_def[7] == nil or next(space_def[7]) == nil then\n"
"        local format = {}\n"
"        format={}\n"
"        format[1] = {name='grantor', type='num'}\n"
"        format[2] = {name='grantee', type='num'}\n"
"        format[3] = {name='object_type', type='str'}\n"
"        format[4] = {name='object_id', type='num'}\n"
"        format[5] = {name='privilege', type='num'}\n"
"        log.info(\"alter space _priv set format\")\n"
"        box.space._priv:format(format)\n"
"    end\n"
"\n"
"    local space_def = box.space._space:get(box.space._cluster.id)\n"
"    if space_def[7] == nil or next(space_def[7]) == nil then\n"
"        local format = {}\n"
"        format[1] = {name='id', type='num'}\n"
"        format[2] = {name='uuid', type='str'}\n"
"        log.info(\"alter space _schema set format\")\n"
"        box.space._cluster:format(format)\n"
"    end\n"
"\n"
"    local spaces = {}\n"
"    for _, space_def in box.space._space:pairs() do\n"
"        if space_def[7] == nil then\n"
"            table.insert(spaces, space_def)\n"
"        end\n"
"    end\n"
"    for _, space_def in ipairs(spaces) do\n"
"        log.info(\"alter space %s set format\", space_def[3])\n"
"        box.space._space:update(space_def[1], {{'=', 7, {}}})\n"
"    end\n"
"end\n"
"\n"
"local function create_sysview(source_id, target_id)\n"
"    --\n"
"    -- Create definitions for the system view, and grant\n"
"    -- privileges on system views to 'PUBLIC' role\n"
"    --\n"
"    local def = box.space._space:get(source_id):totable()\n"
"    def[1] = target_id\n"
"    def[3] = \"_v\"..def[3]:sub(2)\n"
"    def[4] = 'sysview'\n"
"    local space_def = box.space._space:get(target_id)\n"
"    if space_def == nil then\n"
"        log.info(\"create view %s...\", def[3])\n"
"        box.space._space:replace(def)\n"
"    elseif json.encode(space_def[7]) ~= json.encode(def[7]) then\n"
"        -- sync box.space._vXXX format with box.space._XXX format\n"
"        log.info(\"alter space %s set format\", def[3])\n"
"        box.space._space:update(def[1], {{ '=', 7, def[7] }})\n"
"    end\n"
"    local idefs = {}\n"
"    for _, idef in box.space._index:pairs(source_id, { iterator = 'EQ'}) do\n"
"        idef = idef:totable()\n"
"        idef[1] = target_id\n"
"        table.insert(idefs, idef)\n"
"    end\n"
"    for _, idef in ipairs(idefs) do\n"
"        if box.space._index:get({idef[1], idef[2]}) == nil then\n"
"            log.info(\"create index %s on %s\", idef[3], def[3])\n"
"            box.space._index:replace(idef)\n"
"        end\n"
"    end\n"
"    -- public can read system views\n"
"    if box.space._priv.index.primary:count({PUBLIC, 'space', target_id}) == 0 then\n"
"        log.info(\"grant read access to 'public' role for %s view\", def[3])\n"
"        box.space._priv:insert({1, PUBLIC, 'space', target_id, box.priv.R})\n"
"    end\n"
"end\n"
"\n"
"local function upgrade_users_to_1_6_8()\n"
"    if box.space._user.index.name:count({'replication'}) == 0 then\n"
"        log.info(\"create role replication\")\n"
"        local RPL_ID = box.space._user:auto_increment{ADMIN, 'replication', 'role'}[1]\n"
"        -- replication can read the entire universe\n"
"        log.info(\"grant read on universe to replication\")\n"
"        box.space._priv:replace{1, RPL_ID, 'universe', 0, box.priv.R}\n"
"        -- replication can append to '_cluster' system space\n"
"        log.info(\"grant write on space _cluster to replication\")\n"
"        box.space._priv:replace{1, RPL_ID, 'space', box.space._cluster.id, box.priv.W}\n"
"    end\n"
"\n"
"    if box.space._priv.index.primary:count({ADMIN, 'universe', 0}) == 0 then\n"
"        -- grant admin access to the universe\n"
"        log.info(\"grant all on universe to admin\")\n"
"        box.space._priv:insert{ADMIN, ADMIN, 'universe', 0, box.priv.R +\n"
"                                                        box.priv.W + box.priv.X}\n"
"    end\n"
"\n"
"    if box.space._func.index.name:count(\"box.schema.user.info\") == 0 then\n"
"        -- create \"box.schema.user.info\" function\n"
"        log.info('create function \"box.schema.user.info\" with setuid')\n"
"        box.space._func:auto_increment{ADMIN, 'box.schema.user.info', 1}\n"
"\n"
"        -- grant 'public' role access to 'box.schema.user.info' function\n"
"        log.info('grant execute on function \"box.schema.user.info\" to public')\n"
"        box.space._priv:replace{ADMIN, PUBLIC, 'function', 1, box.priv.X}\n"
"    end\n"
"end\n"
"\n"
"local function upgrade_priv_to_1_6_8()\n"
"    -- see e5862c387c7151b812810b1a51086b82a7eedcc4\n"
"    local index_def = box.space._index.index.name:get({312, 'owner'})\n"
"    local parts = index_def[6]\n"
"    if parts[1][1] == 1 then\n"
"        log.info(\"fix index owner for _priv\")\n"
"        parts = {{0, 'num'}}\n"
"        box.space._index:update({index_def[1], index_def[2]},\n"
"                                {{'=', 6, parts }})\n"
"    end\n"
"end\n"
"\n"
"local function upgrade_func_to_1_6_8()\n"
"    local funcs = {}\n"
"    for _, def in box.space._func:pairs() do\n"
"        local new_def = def:totable()\n"
"        if new_def[5] == nil then\n"
"            new_def[5] = 'LUA'\n"
"            table.insert(funcs, new_def)\n"
"        end\n"
"    end\n"
"    for _, def in ipairs(funcs) do\n"
"        box.space._func:update(def[1], {{ '=', 5, def[5] }})\n"
"    end\n"
"end\n"
"\n"
"local function upgrade_to_1_6_8()\n"
"    upgrade_index_options_to_1_6_8()\n"
"    upgrade_space_options_to_1_6_8()\n"
"    upgrade_space_format_to_1_6_8()\n"
"    upgrade_users_to_1_6_8()\n"
"    upgrade_priv_to_1_6_8()\n"
"    upgrade_func_to_1_6_8()\n"
"\n"
"    create_sysview(box.schema.SPACE_ID, box.schema.VSPACE_ID)\n"
"    create_sysview(box.schema.INDEX_ID, box.schema.VINDEX_ID)\n"
"    create_sysview(box.schema.USER_ID, box.schema.VUSER_ID)\n"
"    create_sysview(box.schema.FUNC_ID, box.schema.VFUNC_ID)\n"
"    create_sysview(box.schema.PRIV_ID, box.schema.VPRIV_ID)\n"
"\n"
"    local max_id = box.space._schema:get('max_id')\n"
"    if max_id == nil then\n"
"        local id = box.space._space.index.primary:max()[1]\n"
"        if id < box.schema.SYSTEM_ID_MAX then\n"
"            id = box.schema.SYSTEM_ID_MAX\n"
"        end\n"
"        log.info(\"set max_id to %d\", id)\n"
"        box.space._schema:insert{'max_id', id}\n"
"    end\n"
"end\n"
"\n"
"--------------------------------------------------------------------------------\n"
"-- Tarantool 1.7.1\n"
"--------------------------------------------------------------------------------\n"
"\n"
"local function upgrade_users_to_1_7_1()\n"
"    box.schema.user.passwd('guest', '')\n"
"end\n"
"\n"
"local function upgrade_to_1_7_1()\n"
"    upgrade_users_to_1_7_1()\n"
"end\n"
"\n"
"--------------------------------------------------------------------------------\n"
"-- Tarantool 1.7.2\n"
"--------------------------------------------------------------------------------\n"
"\n"
"local function upgrade_field_types_to_1_7_2()\n"
"    local field_types_v16 = {\n"
"        num = 'unsigned';\n"
"        int = 'integer';\n"
"        str = 'string';\n"
"    };\n"
"    local indexes = {}\n"
"    for _, deftuple in box.space._index:pairs() do\n"
"        local def = deftuple:totable()\n"
"        local changed = false\n"
"        local parts = def[6]\n"
"        for _, part in pairs(parts) do\n"
"            local field_type = part[2]:lower()\n"
"            part[2] = field_types_v16[field_type] or field_type\n"
"            if field_type ~= part[2] then\n"
"                changed = true\n"
"            end\n"
"        end\n"
"        if changed then\n"
"            table.insert(indexes, def)\n"
"        end\n"
"    end\n"
"    for _, new_def in ipairs(indexes) do\n"
"        log.info(\"alter index %s on %s set parts to %s\",\n"
"                 new_def[3], box.space[new_def[1]].name,\n"
"                 json.encode(new_def[6]))\n"
"        box.space._index:replace(new_def)\n"
"    end\n"
"end\n"
"\n"
"local function upgrade_to_1_7_2()\n"
"    upgrade_field_types_to_1_7_2()\n"
"end\n"
"\n"
"--------------------------------------------------------------------------------\n"
"-- Tarantool 1.7.5\n"
"--------------------------------------------------------------------------------\n"
"\n"
"local function create_truncate_space()\n"
"    local _truncate = box.space[box.schema.TRUNCATE_ID]\n"
"\n"
"    log.info(\"create space _truncate\")\n"
"    box.space._space:insert{_truncate.id, ADMIN, '_truncate', 'memtx', 0, setmap({}),\n"
"                            {{name = 'id', type = 'unsigned'}, {name = 'count', type = 'unsigned'}}}\n"
"\n"
"    log.info(\"create index primary on _truncate\")\n"
"    box.space._index:insert{_truncate.id, 0, 'primary', 'tree', {unique = true}, {{0, 'unsigned'}}}\n"
"\n"
"    local _priv = box.space[box.schema.PRIV_ID]\n"
"    _priv:insert{ADMIN, PUBLIC, 'space', _truncate.id, box.priv.W}\n"
"end\n"
"\n"
"local function update_existing_users_to_1_7_5()\n"
"    local def_ids_to_update = {}\n"
"    for _, def in box.space._user:pairs() do\n"
"        local new_def = def:totable()\n"
"        if new_def[5] == nil then\n"
"            table.insert(def_ids_to_update, new_def[1])\n"
"        end\n"
"    end\n"
"    for _, id in ipairs(def_ids_to_update) do\n"
"        box.space._user:update(id, {{'=', 5, setmap({})}})\n"
"    end\n"
"end\n"
"\n"
"local function update_space_formats_to_1_7_5()\n"
"    local format = {}\n"
"    format[1] = {type='string', name='key'}\n"
"    box.space._schema:format(format)\n"
"\n"
"    format = {}\n"
"    format[1] = {name='id', type='unsigned'}\n"
"    format[2] = {name='owner', type='unsigned'}\n"
"    format[3] = {name='name', type='string'}\n"
"    format[4] = {name='engine', type='string'}\n"
"    format[5] = {name='field_count', type='unsigned'}\n"
"    format[6] = {name='flags', type='map'}\n"
"    format[7] = {name='format', type='array'}\n"
"    box.space._space:format(format)\n"
"    box.space._vspace:format(format)\n"
"\n"
"    format = {}\n"
"    format[1] = {name = 'id', type = 'unsigned'}\n"
"    format[2] = {name = 'iid', type = 'unsigned'}\n"
"    format[3] = {name = 'name', type = 'string'}\n"
"    format[4] = {name = 'type', type = 'string'}\n"
"    format[5] = {name = 'opts', type = 'map'}\n"
"    format[6] = {name = 'parts', type = 'array'}\n"
"    box.space._index:format(format)\n"
"    box.space._vindex:format(format)\n"
"\n"
"    format = {}\n"
"    format[1] = {name='id', type='unsigned'}\n"
"    format[2] = {name='owner', type='unsigned'}\n"
"    format[3] = {name='name', type='string'}\n"
"    format[4] = {name='setuid', type='unsigned'}\n"
"    box.space._func:format(format)\n"
"    box.space._vfunc:format(format)\n"
"\n"
"    format = {}\n"
"    format[1] = {name='id', type='unsigned'}\n"
"    format[2] = {name='owner', type='unsigned'}\n"
"    format[3] = {name='name', type='string'}\n"
"    format[4] = {name='type', type='string'}\n"
"    format[5] = {name='auth', type='map'}\n"
"    box.space._user:format(format)\n"
"    box.space._vuser:format(format)\n"
"\n"
"    format = {}\n"
"    format[1] = {name='grantor', type='unsigned'}\n"
"    format[2] = {name='grantee', type='unsigned'}\n"
"    format[3] = {name='object_type', type='string'}\n"
"    format[4] = {name='object_id', type='unsigned'}\n"
"    format[5] = {name='privilege', type='unsigned'}\n"
"    box.space._priv:format(format)\n"
"    box.space._vpriv:format(format)\n"
"\n"
"    format = {}\n"
"    format[1] = {name='id', type='unsigned'}\n"
"    format[2] = {name='uuid', type='string'}\n"
"    box.space._cluster:format(format)\n"
"end\n"
"\n"
"local function upgrade_to_1_7_5()\n"
"    create_truncate_space()\n"
"    update_space_formats_to_1_7_5()\n"
"    update_existing_users_to_1_7_5()\n"
"end\n"
"\n"
"local function initial_1_7_5()\n"
"    -- stick to the following convention:\n"
"    -- prefer user id (owner id) in field #1\n"
"    -- prefer object name in field #2\n"
"    -- index on owner id is index #1\n"
"    -- index on object name is index #2\n"
"    --\n"
"\n"
"    local _schema = box.space[box.schema.SCHEMA_ID]\n"
"    local _space = box.space[box.schema.SPACE_ID]\n"
"    local _index = box.space[box.schema.INDEX_ID]\n"
"    local _func = box.space[box.schema.FUNC_ID]\n"
"    local _user = box.space[box.schema.USER_ID]\n"
"    local _priv = box.space[box.schema.PRIV_ID]\n"
"    local _cluster = box.space[box.schema.CLUSTER_ID]\n"
"    local _truncate = box.space[box.schema.TRUNCATE_ID]\n"
"    local MAP = setmap({})\n"
"\n"
"    --\n"
"    -- _schema\n"
"    --\n"
"    log.info(\"create space _schema\")\n"
"    local format = {}\n"
"    format[1] = {type='string', name='key'}\n"
"    _space:insert{_schema.id, ADMIN, '_schema', 'memtx', 0, MAP, format}\n"
"    log.info(\"create index primary on _schema\")\n"
"    _index:insert{_schema.id, 0, 'primary', 'tree', { unique = true }, {{0, 'string'}}}\n"
"\n"
"    --\n"
"    -- _space\n"
"    --\n"
"    log.info(\"create space _space\")\n"
"    format = {}\n"
"    format[1] = {name='id', type='unsigned'}\n"
"    format[2] = {name='owner', type='unsigned'}\n"
"    format[3] = {name='name', type='string'}\n"
"    format[4] = {name='engine', type='string'}\n"
"    format[5] = {name='field_count', type='unsigned'}\n"
"    format[6] = {name='flags', type='map'}\n"
"    format[7] = {name='format', type='array'}\n"
"    _space:insert{_space.id, ADMIN, '_space', 'memtx', 0, MAP, format}\n"
"    -- space name is unique\n"
"    log.info(\"create index primary on _space\")\n"
"    _index:insert{_space.id, 0, 'primary', 'tree', { unique = true }, {{0, 'unsigned'}}}\n"
"    log.info(\"create index owner on _space\")\n"
"    _index:insert{_space.id, 1, 'owner', 'tree', {unique = false }, {{1, 'unsigned'}}}\n"
"    log.info(\"create index index name on _space\")\n"
"    _index:insert{_space.id, 2, 'name', 'tree', { unique = true }, {{2, 'string'}}}\n"
"    create_sysview(box.schema.SPACE_ID, box.schema.VSPACE_ID)\n"
"\n"
"    --\n"
"    -- _index\n"
"    --\n"
"    log.info(\"create space _index\")\n"
"    format = {}\n"
"    format[1] = {name = 'id', type = 'unsigned'}\n"
"    format[2] = {name = 'iid', type = 'unsigned'}\n"
"    format[3] = {name = 'name', type = 'string'}\n"
"    format[4] = {name = 'type', type = 'string'}\n"
"    format[5] = {name = 'opts', type = 'map'}\n"
"    format[6] = {name = 'parts', type = 'array'}\n"
"    _space:insert{_index.id, ADMIN, '_index', 'memtx', 0, MAP, format}\n"
"    -- index name is unique within a space\n"
"    log.info(\"create index primary on _index\")\n"
"    _index:insert{_index.id, 0, 'primary', 'tree', {unique = true}, {{0, 'unsigned'}, {1, 'unsigned'}}}\n"
"    log.info(\"create index name on _index\")\n"
"    _index:insert{_index.id, 2, 'name', 'tree', {unique = true}, {{0, 'unsigned'}, {2, 'string'}}}\n"
"    create_sysview(box.schema.INDEX_ID, box.schema.VINDEX_ID)\n"
"\n"
"    --\n"
"    -- _func\n"
"    --\n"
"    log.info(\"create space _func\")\n"
"    format = {}\n"
"    format[1] = {name='id', type='unsigned'}\n"
"    format[2] = {name='owner', type='unsigned'}\n"
"    format[3] = {name='name', type='string'}\n"
"    format[4] = {name='setuid', type='unsigned'}\n"
"    _space:insert{_func.id, ADMIN, '_func', 'memtx', 0, MAP, format}\n"
"    -- function name and id are unique\n"
"    log.info(\"create index _func:primary\")\n"
"    _index:insert{_func.id, 0, 'primary', 'tree', {unique = true}, {{0, 'unsigned'}}}\n"
"    log.info(\"create index _func:owner\")\n"
"    _index:insert{_func.id, 1, 'owner', 'tree', {unique = false}, {{1, 'unsigned'}}}\n"
"    log.info(\"create index _func:name\")\n"
"    _index:insert{_func.id, 2, 'name', 'tree', {unique = true}, {{2, 'string'}}}\n"
"    create_sysview(box.schema.FUNC_ID, box.schema.VFUNC_ID)\n"
"\n"
"    --\n"
"    -- _user\n"
"    --\n"
"    log.info(\"create space _user\")\n"
"    format = {}\n"
"    format[1] = {name='id', type='unsigned'}\n"
"    format[2] = {name='owner', type='unsigned'}\n"
"    format[3] = {name='name', type='string'}\n"
"    format[4] = {name='type', type='string'}\n"
"    format[5] = {name='auth', type='map'}\n"
"    _space:insert{_user.id, ADMIN, '_user', 'memtx', 0, MAP, format}\n"
"    -- user name and id are unique\n"
"    log.info(\"create index _func:primary\")\n"
"    _index:insert{_user.id, 0, 'primary', 'tree', {unique = true}, {{0, 'unsigned'}}}\n"
"    log.info(\"create index _func:owner\")\n"
"    _index:insert{_user.id, 1, 'owner', 'tree', {unique = false}, {{1, 'unsigned'}}}\n"
"    log.info(\"create index _func:name\")\n"
"    _index:insert{_user.id, 2, 'name', 'tree', {unique = true}, {{2, 'string'}}}\n"
"    create_sysview(box.schema.USER_ID, box.schema.VUSER_ID)\n"
"\n"
"    --\n"
"    -- _priv\n"
"    --\n"
"    log.info(\"create space _priv\")\n"
"    format = {}\n"
"    format[1] = {name='grantor', type='unsigned'}\n"
"    format[2] = {name='grantee', type='unsigned'}\n"
"    format[3] = {name='object_type', type='string'}\n"
"    format[4] = {name='object_id', type='unsigned'}\n"
"    format[5] = {name='privilege', type='unsigned'}\n"
"    _space:insert{_priv.id, ADMIN, '_priv', 'memtx', 0, MAP, format}\n"
"    -- user id, object type and object id are unique\n"
"    log.info(\"create index primary on _priv\")\n"
"    _index:insert{_priv.id, 0, 'primary', 'tree', {unique = true}, {{1, 'unsigned'}, {2, 'string'}, {3, 'unsigned'}}}\n"
"    -- owner index  - to quickly find all privileges granted by a user\n"
"    log.info(\"create index owner on _priv\")\n"
"    _index:insert{_priv.id, 1, 'owner', 'tree', {unique = false}, {{0, 'unsigned'}}}\n"
"    -- object index - to quickly find all grants on a given object\n"
"    log.info(\"create index object on _priv\")\n"
"    _index:insert{_priv.id, 2, 'object', 'tree', {unique = false}, {{2, 'string'}, {3, 'unsigned'}}}\n"
"    create_sysview(box.schema.PRIV_ID, box.schema.VPRIV_ID)\n"
"\n"
"    --\n"
"    -- _cluster\n"
"    --\n"
"    log.info(\"create space _cluster\")\n"
"    format = {}\n"
"    format[1] = {name='id', type='unsigned'}\n"
"    format[2] = {name='uuid', type='string'}\n"
"    _space:insert{_cluster.id, ADMIN, '_cluster', 'memtx', 0, MAP, format}\n"
"    -- primary key: node id\n"
"    log.info(\"create index primary on _cluster\")\n"
"    _index:insert{_cluster.id, 0, 'primary', 'tree', {unique = true}, {{0, 'unsigned'}}}\n"
"    -- node uuid key: node uuid\n"
"    log.info(\"create index uuid on _cluster\")\n"
"    _index:insert{_cluster.id, 1, 'uuid', 'tree', {unique = true}, {{1, 'string'}}}\n"
"\n"
"    --\n"
"    -- _truncate\n"
"    --\n"
"    log.info(\"create space _truncate\")\n"
"    format = {}\n"
"    format[1] = {name='id', type='unsigned'}\n"
"    format[2] = {name='count', type='unsigned'}\n"
"    _space:insert{_truncate.id, ADMIN, '_truncate', 'memtx', 0, MAP, format}\n"
"    -- primary key: space id\n"
"    log.info(\"create index primary on _truncate\")\n"
"    _index:insert{_truncate.id, 0, 'primary', 'tree', {unique = true}, {{0, 'unsigned'}}}\n"
"\n"
"    --\n"
"    -- Create users\n"
"    --\n"
"    log.info(\"create user guest\")\n"
"    _user:insert{GUEST, ADMIN, 'guest', 'user', MAP}\n"
"    box.schema.user.passwd('guest', '')\n"
"    log.info(\"create user admin\")\n"
"    _user:insert{ADMIN, ADMIN, 'admin', 'user', MAP}\n"
"    log.info(\"create role public\")\n"
"    _user:insert{PUBLIC, ADMIN, 'public', 'role', MAP}\n"
"    log.info(\"create role replication\")\n"
"    _user:insert{REPLICATION, ADMIN, 'replication', 'role', MAP}\n"
"\n"
"    --\n"
"    -- Create grants\n"
"    --\n"
"    log.info(\"grant read,write,execute on universe to admin\")\n"
"    _priv:insert{ADMIN, ADMIN, 'universe', 0, box.priv.R + box.priv.W + box.priv.X}\n"
"\n"
"    -- grant role 'public' to 'guest'\n"
"    log.info(\"grant role public to guest\")\n"
"    _priv:insert{ADMIN, GUEST, 'role', PUBLIC, box.priv.X}\n"
"\n"
"    -- replication can read the entire universe\n"
"    log.info(\"grant read on universe to replication\")\n"
"    _priv:replace{ADMIN, REPLICATION, 'universe', 0, box.priv.R}\n"
"    -- replication can append to '_cluster' system space\n"
"    log.info(\"grant write on space _cluster to replication\")\n"
"    _priv:replace{ADMIN, REPLICATION, 'space', _cluster.id, box.priv.W}\n"
"\n"
"    _priv:insert{ADMIN, PUBLIC, 'space', _truncate.id, box.priv.W}\n"
"\n"
"    -- create \"box.schema.user.info\" function\n"
"    log.info('create function \"box.schema.user.info\" with setuid')\n"
"    _func:replace{1, ADMIN, 'box.schema.user.info', 1, 'LUA'}\n"
"\n"
"    -- grant 'public' role access to 'box.schema.user.info' function\n"
"    log.info('grant execute on function \"box.schema.user.info\" to public')\n"
"    _priv:replace{ADMIN, PUBLIC, 'function', 1, box.priv.X}\n"
"\n"
"    log.info(\"set max_id to box.schema.SYSTEM_ID_MAX\")\n"
"    _schema:insert{'max_id', box.schema.SYSTEM_ID_MAX}\n"
"\n"
"    log.info(\"set schema version to 1.7.5\")\n"
"    _schema:insert({'version', 1, 7, 5})\n"
"end\n"
"\n"
"local sequence_format = {{name = 'id', type = 'unsigned'},\n"
"                         {name = 'owner', type = 'unsigned'},\n"
"                         {name = 'name', type = 'string'},\n"
"                         {name = 'step', type = 'integer'},\n"
"                         {name = 'min', type = 'integer'},\n"
"                         {name = 'max', type = 'integer'},\n"
"                         {name = 'start', type = 'integer'},\n"
"                         {name = 'cache', type = 'integer'},\n"
"                         {name = 'cycle', type = 'boolean'}}\n"
"--------------------------------------------------------------------------------\n"
"-- Tarantool 1.7.6\n"
"local function create_sequence_space()\n"
"    local _space = box.space[box.schema.SPACE_ID]\n"
"    local _index = box.space[box.schema.INDEX_ID]\n"
"    local _sequence = box.space[box.schema.SEQUENCE_ID]\n"
"    local _sequence_data = box.space[box.schema.SEQUENCE_DATA_ID]\n"
"    local _space_sequence = box.space[box.schema.SPACE_SEQUENCE_ID]\n"
"    local MAP = setmap({})\n"
"\n"
"    log.info(\"create space _sequence\")\n"
"    _space:insert{_sequence.id, ADMIN, '_sequence', 'memtx', 0, MAP, sequence_format}\n"
"    log.info(\"create index _sequence:primary\")\n"
"    _index:insert{_sequence.id, 0, 'primary', 'tree', {unique = true}, {{0, 'unsigned'}}}\n"
"    log.info(\"create index _sequence:owner\")\n"
"    _index:insert{_sequence.id, 1, 'owner', 'tree', {unique = false}, {{1, 'unsigned'}}}\n"
"    log.info(\"create index _sequence:name\")\n"
"    _index:insert{_sequence.id, 2, 'name', 'tree', {unique = true}, {{2, 'string'}}}\n"
"\n"
"    log.info(\"create space _sequence_data\")\n"
"    _space:insert{_sequence_data.id, ADMIN, '_sequence_data', 'memtx', 0, MAP,\n"
"                  {{name = 'id', type = 'unsigned'}, {name = 'value', type = 'integer'}}}\n"
"    log.info(\"create index primary on _sequence_data\")\n"
"    _index:insert{_sequence_data.id, 0, 'primary', 'hash', {unique = true}, {{0, 'unsigned'}}}\n"
"\n"
"    log.info(\"create space _space_sequence\")\n"
"    _space:insert{_space_sequence.id, ADMIN, '_space_sequence', 'memtx', 0, MAP,\n"
"                  {{name = 'id', type = 'unsigned'},\n"
"                   {name = 'sequence_id', type = 'unsigned'},\n"
"                   {name = 'is_generated', type = 'boolean'}}}\n"
"    log.info(\"create index _space_sequence:primary\")\n"
"    _index:insert{_space_sequence.id, 0, 'primary', 'tree', {unique = true}, {{0, 'unsigned'}}}\n"
"    log.info(\"create index _space_sequence:sequence\")\n"
"    _index:insert{_space_sequence.id, 1, 'sequence', 'tree', {unique = false}, {{1, 'unsigned'}}}\n"
"end\n"
"\n"
"local function create_collation_space()\n"
"    local _collation = box.space[box.schema.COLLATION_ID]\n"
"\n"
"    log.info(\"create space _collation\")\n"
"    box.space._space:insert{_collation.id, ADMIN, '_collation', 'memtx', 0, setmap({}),\n"
"        { { name = 'id', type = 'unsigned' }, { name = 'name', type = 'string' },\n"
"          { name = 'owner', type = 'unsigned' }, { name = 'type', type = 'string' },\n"
"          { name = 'locale', type = 'string' }, { name = 'opts', type = 'map' } } }\n"
"\n"
"    log.info(\"create index primary on _collation\")\n"
"    box.space._index:insert{_collation.id, 0, 'primary', 'tree', {unique = true}, {{0, 'unsigned'}}}\n"
"\n"
"    log.info(\"create index name on _collation\")\n"
"    box.space._index:insert{_collation.id, 1, 'name', 'tree', {unique = true}, {{1, 'string'}}}\n"
"\n"
"    log.info(\"create predefined collations\")\n"
"    box.space._collation:replace{1, \"unicode\", ADMIN, \"ICU\", \"\", setmap{}}\n"
"    box.space._collation:replace{2, \"unicode_ci\", ADMIN, \"ICU\", \"\", {strength='primary'}}\n"
"\n"
"    local _priv = box.space[box.schema.PRIV_ID]\n"
"    _priv:insert{ADMIN, PUBLIC, 'space', _collation.id, box.priv.W}\n"
"end\n"
"\n"
"local function upgrade_to_1_7_6()\n"
"    create_sequence_space()\n"
"    create_collation_space()\n"
"    -- Trigger space format checking by updating version in _schema.\n"
"end\n"
"\n"
"--------------------------------------------------------------------------------\n"
"--- Tarantool 1.7.7\n"
"--------------------------------------------------------------------------------\n"
"local function upgrade_to_1_7_7()\n"
"    local _priv = box.space[box.schema.PRIV_ID]\n"
"    local _user = box.space[box.schema.USER_ID]\n"
"    --\n"
"    -- grant 'session' and 'usage' to all existing users\n"
"    --\n"
"    for _, v in _user:pairs() do\n"
"        if v[4] ~= \"role\" then\n"
"            _priv:upsert({ADMIN, v[1], \"universe\", 0, box.priv.S + box.priv.U},\n"
"                                                {{\"|\", 5, box.priv.S + box.priv.U}})\n"
"        end\n"
"    end\n"
"    --\n"
"    -- grant 'create' to all users with 'read' and 'write'\n"
"    -- on the universe, since going forward we will require\n"
"    -- 'create' rather than 'read,write' to be able to create\n"
"    -- objects\n"
"    --\n"
"    for _, v in _priv.index.object:pairs{'universe'} do\n"
"        if bit.band(v[5], 1) ~= 0 and bit.band(v[5], 2) ~= 0 then\n"
"            _priv:update({v[2], v[3], v[4]}, {{ \"|\", 5, box.priv.C}})\n"
"        end\n"
"    end\n"
"    -- grant admin all new privileges (session, usage, grant option,\n"
"    -- create, alter, drop and anything that might come up in the future\n"
"    --\n"
"    _priv:upsert({ADMIN, ADMIN, 'universe', 0, box.priv.ALL},\n"
"                 {{ \"|\", 5, box.priv.ALL}})\n"
"    --\n"
"    -- create role 'super' and grant it all privileges on universe\n"
"    --\n"
"    _user:replace{SUPER, ADMIN, 'super', 'role', setmap({})}\n"
"    _priv:replace({ADMIN, SUPER, 'universe', 0, 4294967295})\n"
"end\n"
"\n"
"--------------------------------------------------------------------------------\n"
"--- Tarantool 1.10.0\n"
"--------------------------------------------------------------------------------\n"
"local function create_vsequence_space()\n"
"    create_sysview(box.schema.SEQUENCE_ID, box.schema.VSEQUENCE_ID)\n"
"    box.space._vsequence:format(sequence_format)\n"
"end\n"
"\n"
"local function upgrade_to_1_10_0()\n"
"    create_vsequence_space()\n"
"end\n"
"\n"
"--------------------------------------------------------------------------------\n"
"--- Tarantool 1.10.2\n"
"--------------------------------------------------------------------------------\n"
"local function upgrade_priv_to_1_10_2()\n"
"    local _priv = box.space._priv\n"
"    local _vpriv = box.space._vpriv\n"
"    local format = _priv:format()\n"
"\n"
"    format[4].type = 'scalar'\n"
"    _priv:format(format)\n"
"    format = _vpriv:format()\n"
"    format[4].type = 'scalar'\n"
"    _vpriv:format(format)\n"
"    _priv.index.primary:alter{parts={2, 'unsigned', 3, 'string', 4, 'scalar'}}\n"
"    _vpriv.index.primary:alter{parts={2, 'unsigned', 3, 'string', 4, 'scalar'}}\n"
"    _priv.index.object:alter{parts={3, 'string', 4, 'scalar'}}\n"
"    _vpriv.index.object:alter{parts={3, 'string', 4, 'scalar'}}\n"
"end\n"
"\n"
"local function create_vinyl_deferred_delete_space()\n"
"    local _space = box.space[box.schema.SPACE_ID]\n"
"    local _vinyl_deferred_delete = box.space[box.schema.VINYL_DEFERRED_DELETE_ID]\n"
"\n"
"    local format = {}\n"
"    format[1] = {name = 'space_id', type = 'unsigned'}\n"
"    format[2] = {name = 'lsn', type = 'unsigned'}\n"
"    format[3] = {name = 'tuple', type = 'array'}\n"
"\n"
"    log.info(\"create space _vinyl_deferred_delete\")\n"
"    _space:insert{_vinyl_deferred_delete.id, ADMIN, '_vinyl_deferred_delete',\n"
"                  'blackhole', 0, {group_id = 1}, format}\n"
"end\n"
"\n"
"local function upgrade_to_1_10_2()\n"
"    upgrade_priv_to_1_10_2()\n"
"    create_vinyl_deferred_delete_space()\n"
"end\n"
"\n"
"local function get_version()\n"
"    local version = box.space._schema:get{'version'}\n"
"    if version == nil then\n"
"        error('Missing \"version\" in box.space._schema')\n"
"    end\n"
"    local major = version[2]\n"
"    local minor = version[3]\n"
"    local patch = version[4] or 0\n"
"\n"
"    return mkversion(major, minor, patch)\n"
"end\n"
"\n"
"local function upgrade(options)\n"
"    options = options or {}\n"
"    setmetatable(options, {__index = {auto = false}})\n"
"\n"
"    local version = get_version()\n"
"\n"
"    local handlers = {\n"
"        {version = mkversion(1, 6, 8), func = upgrade_to_1_6_8, auto = false},\n"
"        {version = mkversion(1, 7, 1), func = upgrade_to_1_7_1, auto = false},\n"
"        {version = mkversion(1, 7, 2), func = upgrade_to_1_7_2, auto = false},\n"
"        {version = mkversion(1, 7, 5), func = upgrade_to_1_7_5, auto = true},\n"
"        {version = mkversion(1, 7, 6), func = upgrade_to_1_7_6, auto = false},\n"
"        {version = mkversion(1, 7, 7), func = upgrade_to_1_7_7, auto = true},\n"
"        {version = mkversion(1, 10, 0), func = upgrade_to_1_10_0, auto = true},\n"
"        {version = mkversion(1, 10, 2), func = upgrade_to_1_10_2, auto = true},\n"
"    }\n"
"\n"
"    for _, handler in ipairs(handlers) do\n"
"        if version >= handler.version then\n"
"            goto continue\n"
"        end\n"
"        if options.auto and not handler.auto then\n"
"            log.warn(\"cannot auto upgrade schema version to %s, \" ..\n"
"                     \"please call box.schema.upgrade() manually\",\n"
"                     handler.version)\n"
"            return\n"
"        end\n"
"        handler.func()\n"
"        log.info(\"set schema version to %s\", handler.version)\n"
"        box.space._schema:replace({'version',\n"
"                                   handler.version.major,\n"
"                                   handler.version.minor,\n"
"                                   handler.version.patch})\n"
"        ::continue::\n"
"    end\n"
"end\n"
"\n"
"local function bootstrap()\n"
"    set_system_triggers(false)\n"
"    local version = get_version()\n"
"\n"
"    -- Initial() creates a spaces with 1.6.0 format, but 1.7.6\n"
"    -- checks space formats, that fails initial(). It is because\n"
"    -- bootstrap() is called after box.cfg{}. If box.cfg{} is run\n"
"    -- on 1.7.6, then spaces in the cache contains new 1.7.6\n"
"    -- formats (gh-2754). Spaces in the cache are not updated on\n"
"    -- erase(), because system triggers are turned off.\n"
"\n"
"    -- erase current schema\n"
"    erase()\n"
"    -- insert initial schema\n"
"    if version < mkversion(1, 7, 6) then\n"
"        initial()\n"
"    else\n"
"        initial_1_7_5()\n"
"    end\n"
"\n"
"    -- upgrade schema to the latest version\n"
"    upgrade()\n"
"\n"
"    set_system_triggers(true)\n"
"\n"
"    -- save new bootstrap.snap\n"
"    box.snapshot()\n"
"end\n"
"\n"
"box.schema.upgrade = upgrade;\n"
"box.internal.bootstrap = bootstrap;\n"
""
;
