const char buffer_lua[] =
"-- buffer.lua (internal file)\n"
"\n"
"local ffi = require('ffi')\n"
"local READAHEAD = 16320\n"
"\n"
"ffi.cdef[[\n"
"struct slab_cache;\n"
"struct slab_cache *\n"
"tarantool_lua_slab_cache();\n"
"extern struct ibuf *tarantool_lua_ibuf;\n"
"\n"
"struct ibuf\n"
"{\n"
"    struct slab_cache *slabc;\n"
"    char *buf;\n"
"    /** Start of input. */\n"
"    char *rpos;\n"
"    /** End of useful input */\n"
"    char *wpos;\n"
"    /** End of ibuf. */\n"
"    char *epos;\n"
"    size_t start_capacity;\n"
"};\n"
"\n"
"void\n"
"ibuf_create(struct ibuf *ibuf, struct slab_cache *slabc, size_t start_capacity);\n"
"\n"
"void\n"
"ibuf_destroy(struct ibuf *ibuf);\n"
"\n"
"void\n"
"ibuf_reinit(struct ibuf *ibuf);\n"
"\n"
"void *\n"
"ibuf_reserve_slow(struct ibuf *ibuf, size_t size);\n"
"]]\n"
"\n"
"local builtin = ffi.C\n"
"local ibuf_t = ffi.typeof('struct ibuf')\n"
"\n"
"local function errorf(s, ...)\n"
"    error(string.format(s, ...))\n"
"end\n"
"\n"
"local function checkibuf(buf, method)\n"
"    if not ffi.istype(ibuf_t, buf) then\n"
"        errorf('Attempt to call method without object, use ibuf:%s()', method)\n"
"    end\n"
"end\n"
"\n"
"local function ibuf_capacity(buf)\n"
"    checkibuf(buf, 'capacity')\n"
"    return tonumber(buf.epos - buf.buf)\n"
"end\n"
"\n"
"local function ibuf_pos(buf)\n"
"    checkibuf(buf, 'pos')\n"
"    return tonumber(buf.rpos - buf.buf)\n"
"end\n"
"\n"
"local function ibuf_used(buf)\n"
"    checkibuf(buf, 'size')\n"
"    return tonumber(buf.wpos - buf.rpos)\n"
"end\n"
"\n"
"local function ibuf_unused(buf)\n"
"    checkibuf(buf, 'unused')\n"
"    return tonumber(buf.epos - buf.wpos)\n"
"end\n"
"\n"
"local function ibuf_recycle(buf)\n"
"    checkibuf(buf, 'recycle')\n"
"    builtin.ibuf_reinit(buf)\n"
"end\n"
"\n"
"local function ibuf_reset(buf)\n"
"    checkibuf(buf, 'reset')\n"
"    buf.rpos = buf.buf\n"
"    buf.wpos = buf.buf\n"
"end\n"
"\n"
"local function ibuf_reserve_slow(buf, size)\n"
"    local ptr = builtin.ibuf_reserve_slow(buf, size)\n"
"    if ptr == nil then\n"
"        errorf(\"Failed to allocate %d bytes in ibuf\", size)\n"
"    end\n"
"    return ffi.cast('char *', ptr)\n"
"end\n"
"\n"
"local function ibuf_reserve(buf, size)\n"
"    checkibuf(buf, 'reserve')\n"
"    if buf.wpos + size <= buf.epos then\n"
"        return buf.wpos\n"
"    end\n"
"    return ibuf_reserve_slow(buf, size)\n"
"end\n"
"\n"
"local function ibuf_alloc(buf, size)\n"
"    checkibuf(buf, 'alloc')\n"
"    local wpos\n"
"    if buf.wpos + size <= buf.epos then\n"
"        wpos = buf.wpos\n"
"    else\n"
"        wpos = ibuf_reserve_slow(buf, size)\n"
"    end\n"
"    buf.wpos = buf.wpos + size\n"
"    return wpos\n"
"end\n"
"\n"
"local function checksize(buf, size)\n"
"    if buf.rpos + size > buf.wpos then\n"
"        errorf(\"Attempt to read out of range bytes: needed=%d size=%d\",\n"
"            tonumber(size), ibuf_used(buf))\n"
"    end\n"
"end\n"
"\n"
"local function ibuf_checksize(buf, size)\n"
"    checkibuf(buf, 'checksize')\n"
"    checksize(buf, size)\n"
"    return buf.rpos\n"
"end\n"
"\n"
"local function ibuf_read(buf, size)\n"
"    checkibuf(buf, 'read')\n"
"    checksize(buf, size)\n"
"    local rpos = buf.rpos\n"
"    buf.rpos = rpos + size\n"
"    return rpos\n"
"end\n"
"\n"
"local function ibuf_serialize(buf)\n"
"    local properties = { rpos = buf.rpos, wpos = buf.wpos }\n"
"    return { ibuf = properties }\n"
"end\n"
"\n"
"local ibuf_methods = {\n"
"    recycle = ibuf_recycle;\n"
"    reset = ibuf_reset;\n"
"\n"
"    reserve = ibuf_reserve;\n"
"    alloc = ibuf_alloc;\n"
"\n"
"    checksize = ibuf_checksize;\n"
"    read = ibuf_read;\n"
"    __serialize = ibuf_serialize;\n"
"\n"
"    size = ibuf_used;\n"
"    capacity = ibuf_capacity;\n"
"    pos = ibuf_pos;\n"
"    unused = ibuf_unused;\n"
"}\n"
"\n"
"local function ibuf_tostring(ibuf)\n"
"    return '<ibuf>'\n"
"end\n"
"local ibuf_mt = {\n"
"    __gc = ibuf_recycle;\n"
"    __index = ibuf_methods;\n"
"    __tostring = ibuf_tostring;\n"
"};\n"
"\n"
"ffi.metatype(ibuf_t, ibuf_mt);\n"
"\n"
"local function ibuf_new(arg, arg2)\n"
"    local buf = ffi.new(ibuf_t)\n"
"    local slabc = builtin.tarantool_lua_slab_cache()\n"
"    builtin.ibuf_create(buf, slabc, READAHEAD)\n"
"    if arg == nil then\n"
"        return buf\n"
"    elseif type(arg) == 'number' then\n"
"        ibuf_reserve(buf, arg)\n"
"        return buf\n"
"    end\n"
"    errorf('Usage: ibuf([size])')\n"
"end\n"
"\n"
"return {\n"
"    ibuf = ibuf_new;\n"
"    IBUF_SHARED = ffi.C.tarantool_lua_ibuf;\n"
"    READAHEAD = READAHEAD;\n"
"}\n"
""
;
