const char crypto_lua[] =
"-- crypto.lua (internal file)\n"
"\n"
"local ffi = require('ffi')\n"
"local buffer = require('buffer')\n"
"\n"
"ffi.cdef[[\n"
"    int tnt_openssl_init(void);\n"
"    /* from openssl/err.h */\n"
"    unsigned long ERR_get_error(void);\n"
"    char *ERR_error_string(unsigned long e, char *buf);\n"
"\n"
"    /* from openssl/evp.h */\n"
"    typedef void ENGINE;\n"
"\n"
"    typedef struct {} EVP_MD_CTX;\n"
"    typedef struct {} EVP_MD;\n"
"    EVP_MD_CTX *tnt_EVP_MD_CTX_new(void);\n"
"    void tnt_EVP_MD_CTX_free(EVP_MD_CTX *ctx);\n"
"    int EVP_DigestInit_ex(EVP_MD_CTX *ctx, const EVP_MD *type, ENGINE *impl);\n"
"    int EVP_DigestUpdate(EVP_MD_CTX *ctx, const void *d, size_t cnt);\n"
"    int EVP_DigestFinal_ex(EVP_MD_CTX *ctx, unsigned char *md, unsigned int *s);\n"
"    const EVP_MD *EVP_get_digestbyname(const char *name);\n"
"\n"
"    typedef struct {} HMAC_CTX;\n"
"    HMAC_CTX *tnt_HMAC_CTX_new(void);\n"
"    void tnt_HMAC_CTX_free(HMAC_CTX *ctx);\n"
"    int HMAC_Init_ex(HMAC_CTX *ctx, const void *key, int len,\n"
"                 const EVP_MD *md, ENGINE *impl);\n"
"    int HMAC_Update(HMAC_CTX *ctx, const unsigned char *data, size_t len);\n"
"    int HMAC_Final(HMAC_CTX *ctx, unsigned char *md, unsigned int *len);\n"
"\n"
"    typedef struct {} EVP_CIPHER_CTX;\n"
"    typedef struct {} EVP_CIPHER;\n"
"    EVP_CIPHER_CTX *EVP_CIPHER_CTX_new();\n"
"    void EVP_CIPHER_CTX_free(EVP_CIPHER_CTX *ctx);\n"
"\n"
"    int EVP_CipherInit_ex(EVP_CIPHER_CTX *ctx, const EVP_CIPHER *cipher,\n"
"                          ENGINE *impl, const unsigned char *key,\n"
"                          const unsigned char *iv, int enc);\n"
"    int EVP_CipherUpdate(EVP_CIPHER_CTX *ctx, unsigned char *out, int *outl,\n"
"                     const unsigned char *in, int inl);\n"
"    int EVP_CipherFinal_ex(EVP_CIPHER_CTX *ctx, unsigned char *out, int *outl);\n"
"    int EVP_CIPHER_CTX_cleanup(EVP_CIPHER_CTX *ctx);\n"
"\n"
"    int tnt_EVP_CIPHER_iv_length(const EVP_CIPHER *cipher);\n"
"    int tnt_EVP_CIPHER_key_length(const EVP_CIPHER *cipher);\n"
"\n"
"    int EVP_CIPHER_block_size(const EVP_CIPHER *cipher);\n"
"    const EVP_CIPHER *EVP_get_cipherbyname(const char *name);\n"
"]]\n"
"\n"
"ffi.C.tnt_openssl_init();\n"
"\n"
"local function openssl_err_str()\n"
"  return ffi.string(ffi.C.ERR_error_string(ffi.C.ERR_get_error(), nil))\n"
"end\n"
"\n"
"local digests = {}\n"
"for class, name in pairs({\n"
"    md2 = 'MD2', md4 = 'MD4', md5 = 'MD5',\n"
"    sha = 'SHA', sha1 = 'SHA1', sha224 = 'SHA224',\n"
"    sha256 = 'SHA256', sha384 = 'SHA384', sha512 = 'SHA512',\n"
"    dss = 'DSS', dss1 = 'DSS1', mdc2 = 'MDC2', ripemd160 = 'RIPEMD160'}) do\n"
"    local digest = ffi.C.EVP_get_digestbyname(class)\n"
"    if digest ~= nil then\n"
"        digests[class] = digest\n"
"    end\n"
"end\n"
"\n"
"local digest_mt = {}\n"
"\n"
"local function digest_gc(ctx)\n"
"    ffi.C.tnt_EVP_MD_CTX_free(ctx)\n"
"end\n"
"\n"
"local function digest_new(digest)\n"
"    local ctx = ffi.C.tnt_EVP_MD_CTX_new()\n"
"    if ctx == nil then\n"
"        return error('Can\\'t create digest ctx: ' .. openssl_err_str())\n"
"    end\n"
"    ffi.gc(ctx, digest_gc)\n"
"    local self = setmetatable({\n"
"        ctx = ctx,\n"
"        digest = digest,\n"
"        buf = buffer.ibuf(64),\n"
"        initialized = false,\n"
"        outl = ffi.new('int[1]')\n"
"    }, digest_mt)\n"
"    self:init()\n"
"    return self\n"
"end\n"
"\n"
"local function digest_init(self)\n"
"    if self.ctx == nil then\n"
"        return error('Digest context isn\\'t usable')\n"
"    end\n"
"    if ffi.C.EVP_DigestInit_ex(self.ctx, self.digest, nil) ~= 1 then\n"
"        return error('Can\\'t init digest: ' .. openssl_err_str())\n"
"    end\n"
"    self.initialized = true\n"
"end\n"
"\n"
"local function digest_update(self, input)\n"
"    if not self.initialized then\n"
"        return error('Digest not initialized')\n"
"    end\n"
"    if ffi.C.EVP_DigestUpdate(self.ctx, input, input:len()) ~= 1 then\n"
"        return error('Can\\'t update digest: ' .. openssl_err_str())\n"
"    end\n"
"end\n"
"\n"
"local function digest_final(self)\n"
"    if not self.initialized then\n"
"        return error('Digest not initialized')\n"
"    end\n"
"    self.initialized = false\n"
"    if ffi.C.EVP_DigestFinal_ex(self.ctx, self.buf.wpos, self.outl) ~= 1 then\n"
"        return error('Can\\'t finalize digest: ' .. openssl_err_str())\n"
"    end\n"
"    return ffi.string(self.buf.wpos, self.outl[0])\n"
"end\n"
"\n"
"local function digest_free(self)\n"
"    ffi.C.tnt_EVP_MD_CTX_free(self.ctx)\n"
"    ffi.gc(self.ctx, nil)\n"
"    self.ctx = nil\n"
"    self.initialized = false\n"
"end\n"
"\n"
"digest_mt = {\n"
"    __index = {\n"
"          init = digest_init,\n"
"          update = digest_update,\n"
"          result = digest_final,\n"
"          free = digest_free\n"
"    }\n"
"}\n"
"\n"
"local hmacs = digests\n"
"\n"
"local hmac_mt = {}\n"
"\n"
"local function hmac_gc(ctx)\n"
"    ffi.C.tnt_HMAC_CTX_free(ctx)\n"
"end\n"
"\n"
"local function hmac_new(digest, key)\n"
"    if key == nil then\n"
"        return error('Key should be specified for HMAC operations')\n"
"    end\n"
"    local ctx = ffi.C.tnt_HMAC_CTX_new()\n"
"    if ctx == nil then\n"
"        return error('Can\\'t create HMAC ctx: ' .. openssl_err_str())\n"
"    end\n"
"    ffi.gc(ctx, hmac_gc)\n"
"    local self = setmetatable({\n"
"        ctx = ctx,\n"
"        digest = digest,\n"
"        buf = buffer.ibuf(64),\n"
"        initialized = false,\n"
"        outl = ffi.new('int[1]')\n"
"    }, hmac_mt)\n"
"    self:init(key)\n"
"    return self\n"
"end\n"
"\n"
"local function hmac_init(self, key)\n"
"    if self.ctx == nil then\n"
"        return error('HMAC context isn\\'t usable')\n"
"    end\n"
"    if ffi.C.HMAC_Init_ex(self.ctx, key, key:len(), self.digest, nil) ~= 1 then\n"
"        return error('Can\\'t init HMAC: ' .. openssl_err_str())\n"
"    end\n"
"    self.initialized = true\n"
"end\n"
"\n"
"local function hmac_update(self, input)\n"
"    if not self.initialized then\n"
"        return error('HMAC not initialized')\n"
"    end\n"
"    if ffi.C.HMAC_Update(self.ctx, input, input:len()) ~= 1 then\n"
"        return error('Can\\'t update HMAC: ' .. openssl_err_str())\n"
"    end\n"
"end\n"
"\n"
"local function hmac_final(self)\n"
"    if not self.initialized then\n"
"        return error('HMAC not initialized')\n"
"    end\n"
"    self.initialized = false\n"
"    if ffi.C.HMAC_Final(self.ctx, self.buf.wpos, self.outl) ~= 1 then\n"
"        return error('Can\\'t finalize HMAC: ' .. openssl_err_str())\n"
"    end\n"
"    return ffi.string(self.buf.wpos, self.outl[0])\n"
"end\n"
"\n"
"local function hmac_free(self)\n"
"    ffi.C.tnt_HMAC_CTX_free(self.ctx)\n"
"    ffi.gc(self.ctx, nil)\n"
"    self.ctx = nil\n"
"    self.initialized = false\n"
"end\n"
"\n"
"hmac_mt = {\n"
"    __index = {\n"
"          init = hmac_init,\n"
"          update = hmac_update,\n"
"          result = hmac_final,\n"
"          free = hmac_free\n"
"    }\n"
"}\n"
"\n"
"local ciphers = {}\n"
"for algo, algo_name in pairs({des = 'DES', aes128 = 'AES-128',\n"
"    aes192 = 'AES-192', aes256 = 'AES-256'}) do\n"
"    local algo_api = {}\n"
"    for mode, mode_name in pairs({cfb = 'CFB', ofb = 'OFB',\n"
"        cbc = 'CBC', ecb = 'ECB'}) do\n"
"            local cipher =\n"
"                ffi.C.EVP_get_cipherbyname(algo_name .. '-' .. mode_name)\n"
"            if cipher ~= nil then\n"
"                algo_api[mode] = cipher\n"
"            end\n"
"    end\n"
"    if algo_api ~= {} then\n"
"        ciphers[algo] = algo_api\n"
"    end\n"
"end\n"
"\n"
"local cipher_mt = {}\n"
"\n"
"local function cipher_gc(ctx)\n"
"    ffi.C.EVP_CIPHER_CTX_free(ctx)\n"
"end\n"
"\n"
"local function cipher_new(cipher, key, iv, direction)\n"
"    local ctx = ffi.C.EVP_CIPHER_CTX_new()\n"
"    if ctx == nil then\n"
"        return error('Can\\'t create cipher ctx: ' .. openssl_err_str())\n"
"    end\n"
"    ffi.gc(ctx, cipher_gc)\n"
"    local self = setmetatable({\n"
"        ctx = ctx,\n"
"        cipher = cipher,\n"
"        block_size = ffi.C.EVP_CIPHER_block_size(cipher),\n"
"        direction = direction,\n"
"        buf = buffer.ibuf(),\n"
"        initialized = false,\n"
"        outl = ffi.new('int[1]')\n"
"    }, cipher_mt)\n"
"    self:init(key, iv)\n"
"    return self\n"
"end\n"
"\n"
"local function cipher_init(self, key, iv)\n"
"    if self.ctx == nil then\n"
"        return error('Cipher context isn\\'t usable')\n"
"    end\n"
"    local cipher = self.cipher\n"
"    key = key or self.key\n"
"    iv = iv or self.iv\n"
"    local needed = ffi.C.tnt_EVP_CIPHER_key_length(cipher)\n"
"    if key ~= nil and key:len() ~= needed then\n"
"        return error('Key length should be equal to cipher key length ('..\n"
"                     tostring(needed)..' bytes)')\n"
"    end\n"
"    needed = ffi.C.tnt_EVP_CIPHER_iv_length(cipher)\n"
"    if iv ~= nil and iv:len() ~= needed then\n"
"        return error('Initial vector length should be equal to cipher iv '..\n"
"                     'length ('..tostring(needed)..' bytes)')\n"
"    end\n"
"    self.key = key\n"
"    self.iv = iv\n"
"    if key and iv then\n"
"        if ffi.C.EVP_CipherInit_ex(self.ctx, cipher, nil, key, iv,\n"
"                                   self.direction) ~= 1 then\n"
"            return error('Can\\'t init cipher:'..openssl_err_str())\n"
"        end\n"
"        self.initialized = true\n"
"    end\n"
"end\n"
"\n"
"local function cipher_update(self, input)\n"
"    if not self.initialized then\n"
"        return error('Cipher not initialized')\n"
"    end\n"
"    if type(input) ~= 'string' then\n"
"        error(\"Usage: cipher:update(string)\")\n"
"    end\n"
"    local wpos = self.buf:reserve(input:len() + self.block_size - 1)\n"
"    if ffi.C.EVP_CipherUpdate(self.ctx, wpos, self.outl, input, input:len()) ~= 1 then\n"
"        return error('Can\\'t update cipher:' .. openssl_err_str())\n"
"    end\n"
"    return ffi.string(wpos, self.outl[0])\n"
"end\n"
"\n"
"local function cipher_final(self)\n"
"    if not self.initialized then\n"
"        return error('Cipher not initialized')\n"
"    end\n"
"    local wpos = self.buf:reserve(self.block_size)\n"
"    if ffi.C.EVP_CipherFinal_ex(self.ctx, wpos, self.outl) ~= 1 then\n"
"        return error('Can\\'t finalize cipher:' .. openssl_err_str())\n"
"    end\n"
"    return ffi.string(wpos, self.outl[0])\n"
"end\n"
"\n"
"local function cipher_free(self)\n"
"    ffi.C.EVP_CIPHER_CTX_free(self.ctx)\n"
"    ffi.gc(self.ctx, nil)\n"
"    self.ctx = nil\n"
"    self.initialized = false\n"
"    self.buf:reset()\n"
"end\n"
"\n"
"cipher_mt = {\n"
"    __index = {\n"
"          init = cipher_init,\n"
"          update = cipher_update,\n"
"          result = cipher_final,\n"
"          free = cipher_free\n"
"    }\n"
"}\n"
"\n"
"local digest_api = {}\n"
"for class, digest in pairs(digests) do\n"
"    digest_api[class] = setmetatable({\n"
"        new = function () return digest_new(digest) end\n"
"    }, {\n"
"        __call = function (self, str)\n"
"            if type(str) ~= 'string' then\n"
"                error(\"Usage: digest.\"..class..\"(string)\")\n"
"            end\n"
"            local ctx = digest_new(digest)\n"
"            ctx:update(str)\n"
"            local res = ctx:result()\n"
"            ctx:free()\n"
"            return res\n"
"        end\n"
"    })\n"
"end\n"
"\n"
"digest_api = setmetatable(digest_api,\n"
"    {__index = function(self, digest)\n"
"        return error('Digest method \"' .. digest .. '\" is not supported')\n"
"    end })\n"
"\n"
"local hmac_api = {}\n"
"for class, digest in pairs(hmacs) do\n"
"    hmac_api[class] = setmetatable({\n"
"        new = function (key) return hmac_new(digest, key) end\n"
"    }, {\n"
"        __call = function (self, key, str)\n"
"            if type(str) ~= 'string' then\n"
"                error(\"Usage: hmac.\"..class..\"(key, string)\")\n"
"            end\n"
"            local ctx = hmac_new(digest, key)\n"
"            ctx:update(str)\n"
"            local res = ctx:result()\n"
"            ctx:free()\n"
"            return res\n"
"        end\n"
"    })\n"
"    hmac_api[class .. '_hex'] = function (key, str)\n"
"        if type(str) ~= 'string' then\n"
"            error(\"Usage: hmac.\"..class..\"_hex(key, string)\")\n"
"        end\n"
"        return string.hex(hmac_api[class](key, str))\n"
"    end\n"
"end\n"
"\n"
"hmac_api = setmetatable(hmac_api,\n"
"    {__index = function(self, digest)\n"
"        return error('HMAC method \"' .. digest .. '\" is not supported')\n"
"    end })\n"
"\n"
"local function cipher_mode_error(self, mode)\n"
"  error('Cipher mode ' .. mode .. ' is not supported')\n"
"end\n"
"\n"
"local cipher_api = {}\n"
"for class, subclass in pairs(ciphers) do\n"
"    local class_api = {}\n"
"    for subclass, cipher in pairs(subclass) do\n"
"        class_api[subclass] = {}\n"
"        for direction, param in pairs({encrypt = 1, decrypt = 0}) do\n"
"              class_api[subclass][direction] = setmetatable({\n"
"                new = function (key, iv)\n"
"                    return cipher_new(cipher, key, iv, param)\n"
"                end\n"
"            }, {\n"
"                __call = function (self, str, key, iv)\n"
"                    local ctx = cipher_new(cipher, key, iv, param)\n"
"                    local res = ctx:update(str)\n"
"                    res = res .. ctx:result()\n"
"                    ctx:free()\n"
"                    return res\n"
"                end\n"
"            })\n"
"        end\n"
"    end\n"
"    class_api = setmetatable(class_api, {__index = cipher_mode_error})\n"
"    if class_api ~= {} then\n"
"        cipher_api[class] = class_api\n"
"    end\n"
"end\n"
"\n"
"cipher_api = setmetatable(cipher_api,\n"
"    {__index = function(self, cipher)\n"
"        return error('Cipher method \"' .. cipher .. '\" is not supported')\n"
"    end })\n"
"\n"
"return {\n"
"    digest = digest_api,\n"
"    hmac   = hmac_api,\n"
"    cipher = cipher_api,\n"
"}\n"
""
;
