const char init_lua[] =
"-- init.lua -- internal file\n"
"\n"
"local ffi = require('ffi')\n"
"ffi.cdef[[\n"
"struct type_info;\n"
"struct method_info;\n"
"struct error;\n"
"\n"
"enum ctype {\n"
"    CTYPE_VOID = 0,\n"
"    CTYPE_INT,\n"
"    CTYPE_CONST_CHAR_PTR\n"
"};\n"
"\n"
"struct type_info {\n"
"    const char *name;\n"
"    const struct type_info *parent;\n"
"    const struct method_info *methods;\n"
"};\n"
"\n"
"enum {\n"
"    DIAG_ERRMSG_MAX = 512,\n"
"    DIAG_FILENAME_MAX = 256\n"
"};\n"
"\n"
"typedef void (*error_f)(struct error *e);\n"
"\n"
"struct error {\n"
"    error_f _destroy;\n"
"    error_f _raise;\n"
"    error_f _log;\n"
"    const struct type_info *_type;\n"
"    int _refs;\n"
"    /** Line number. */\n"
"    unsigned _line;\n"
"    /* Source file name. */\n"
"    char _file[DIAG_FILENAME_MAX];\n"
"    /* Error description. */\n"
"    char _errmsg[DIAG_ERRMSG_MAX];\n"
"};\n"
"\n"
"enum { METHOD_ARG_MAX = 8 };\n"
"\n"
"struct method_info {\n"
"    const struct type_info *owner;\n"
"    const char *name;\n"
"    enum ctype rtype;\n"
"    enum ctype atype[METHOD_ARG_MAX];\n"
"    int nargs;\n"
"    bool isconst;\n"
"\n"
"    union {\n"
"        /* Add extra space to get proper struct size in C */\n"
"        void *_spacer[2];\n"
"    };\n"
"};\n"
"\n"
"char *\n"
"exception_get_string(struct error *e, const struct method_info *method);\n"
"int\n"
"exception_get_int(struct error *e, const struct method_info *method);\n"
"\n"
"double\n"
"tarantool_uptime(void);\n"
"typedef int32_t pid_t;\n"
"pid_t getpid(void);\n"
"]]\n"
"\n"
"local fio = require(\"fio\")\n"
"\n"
"local REFLECTION_CACHE = {}\n"
"\n"
"local function reflection_enumerate(err)\n"
"    local key = tostring(err._type)\n"
"    local result = REFLECTION_CACHE[key]\n"
"    if result ~= nil then\n"
"        return result\n"
"    end\n"
"    result = {}\n"
"    -- See type_foreach_method() in reflection.h\n"
"    local t = err._type\n"
"    while t ~= nil do\n"
"        local m = t.methods\n"
"        while m.name ~= nil do\n"
"            result[ffi.string(m.name)] = m\n"
"            m = m + 1\n"
"        end\n"
"        t = t.parent\n"
"    end\n"
"    REFLECTION_CACHE[key] = result\n"
"    return result\n"
"end\n"
"\n"
"local function reflection_get(err, method)\n"
"    if method.nargs ~= 0 then\n"
"        return nil -- NYI\n"
"    end\n"
"    if method.rtype == ffi.C.CTYPE_INT then\n"
"        return tonumber(ffi.C.exception_get_int(err, method))\n"
"    elseif method.rtype == ffi.C.CTYPE_CONST_CHAR_PTR then\n"
"        local str = ffi.C.exception_get_string(err, method)\n"
"        if str == nil then\n"
"            return nil\n"
"        end\n"
"        return ffi.string(str)\n"
"    end\n"
"end\n"
"\n"
"local function error_type(err)\n"
"    return ffi.string(err._type.name)\n"
"end\n"
"\n"
"local function error_message(err)\n"
"    return ffi.string(err._errmsg)\n"
"end\n"
"\n"
"local function error_trace(err)\n"
"    if err._file[0] == 0 then\n"
"        return {}\n"
"    end\n"
"    return {\n"
"        { file = ffi.string(err._file), line = tonumber(err._line) };\n"
"    }\n"
"end\n"
"\n"
"local error_fields = {\n"
"    [\"type\"]        = error_type;\n"
"    [\"message\"]     = error_message;\n"
"    [\"trace\"]       = error_trace;\n"
"}\n"
"\n"
"local function error_unpack(err)\n"
"    if not ffi.istype('struct error', err) then\n"
"        error(\"Usage: error:unpack()\")\n"
"    end\n"
"    local result = {}\n"
"    for key, getter in pairs(error_fields)  do\n"
"        result[key] = getter(err)\n"
"    end\n"
"    for key, getter in pairs(reflection_enumerate(err)) do\n"
"        local value = reflection_get(err, getter)\n"
"        if value ~= nil then\n"
"            result[key] = value\n"
"        end\n"
"    end\n"
"    return result\n"
"end\n"
"\n"
"local function error_raise(err)\n"
"    if not ffi.istype('struct error', err) then\n"
"        error(\"Usage: error:raise()\")\n"
"    end\n"
"    error(err)\n"
"end\n"
"\n"
"local function error_match(err, ...)\n"
"    if not ffi.istype('struct error', err) then\n"
"        error(\"Usage: error:match()\")\n"
"    end\n"
"    return string.match(error_message(err), ...)\n"
"end\n"
"\n"
"local function error_serialize(err)\n"
"    -- Return an error message only in admin console to keep compatibility\n"
"    return error_message(err)\n"
"end\n"
"\n"
"local error_methods = {\n"
"    [\"unpack\"] = error_unpack;\n"
"    [\"raise\"] = error_raise;\n"
"    [\"match\"] = error_match; -- Tarantool 1.6 backward compatibility\n"
"    [\"__serialize\"] = error_serialize;\n"
"}\n"
"\n"
"local function error_index(err, key)\n"
"    local getter = error_fields[key]\n"
"    if getter ~= nil then\n"
"        return getter(err)\n"
"    end\n"
"    getter = reflection_enumerate(err)[key]\n"
"    if getter ~= nil and getter.nargs == 0 then\n"
"        local val = reflection_get(err, getter)\n"
"        if val ~= nil then\n"
"            return val\n"
"        end\n"
"    end\n"
"    return error_methods[key]\n"
"end\n"
"\n"
"local function error_concat(lhs, rhs)\n"
"    if ffi.istype('struct error', lhs) then\n"
"        return tostring(lhs) .. rhs\n"
"    elseif ffi.istype('struct error', rhs) then\n"
"        return lhs .. tostring(rhs)\n"
"    else\n"
"       error('error_mt.__concat(): neither of args is an error')\n"
"    end\n"
"end\n"
"\n"
"local error_mt = {\n"
"    __index = error_index;\n"
"    __tostring = error_message;\n"
"    __concat = error_concat;\n"
"};\n"
"\n"
"ffi.metatype('struct error', error_mt);\n"
"\n"
"dostring = function(s, ...)\n"
"    local chunk, message = loadstring(s)\n"
"    if chunk == nil then\n"
"        error(message, 2)\n"
"    end\n"
"    return chunk(...)\n"
"end\n"
"\n"
"local function uptime()\n"
"    return tonumber(ffi.C.tarantool_uptime());\n"
"end\n"
"\n"
"local function pid()\n"
"    return tonumber(ffi.C.getpid())\n"
"end\n"
"\n"
"local soext = (jit.os == \"OSX\" and \"dylib\" or \"so\")\n"
"\n"
"local function mksymname(name)\n"
"    local mark = string.find(name, \"-\")\n"
"    if mark then name = string.sub(name, mark + 1) end\n"
"    return \"luaopen_\" .. string.gsub(name, \"%.\", \"_\")\n"
"end\n"
"\n"
"local function load_lib(file, name)\n"
"    return package.loadlib(file, mksymname(name))\n"
"end\n"
"\n"
"local function load_lua(file)\n"
"    return loadfile(file)\n"
"end\n"
"\n"
"local function search_cwd_lib(name)\n"
"    local path = \"./\?.\"..soext\n"
"    return package.searchpath(name, path)\n"
"end\n"
"\n"
"local function search_cwd_lua(name)\n"
"    local path = \"./\?.lua;./\?/init.lua\"\n"
"    return package.searchpath(name, path)\n"
"end\n"
"\n"
"local function traverse_rocks(name, pathes_search)\n"
"    local cwd = fio.cwd()\n"
"    local index = string.len(cwd) + 1\n"
"    local strerr = \"\"\n"
"    while index ~= nil do\n"
"        cwd = string.sub(cwd, 1, index - 1)\n"
"        for i, path in ipairs(pathes_search) do\n"
"            local file, err = package.searchpath(name, cwd .. path)\n"
"            if err == nil then\n"
"                return file\n"
"            end\n"
"            strerr = strerr .. err\n"
"        end\n"
"        index = string.find(cwd, \"/[^/]*$\")\n"
"    end\n"
"    return nil, strerr\n"
"end\n"
"\n"
"local function search_rocks_lua(name)\n"
"    local pathes_search = {\n"
"        \"/.rocks/share/tarantool/\?.lua;\",\n"
"        \"/.rocks/share/tarantool/\?/init.lua;\",\n"
"    }\n"
"    return traverse_rocks(name, pathes_search)\n"
"end\n"
"\n"
"local function search_rocks_lib(name)\n"
"    local pathes_search = {\n"
"        \"/.rocks/lib/tarantool/\?.\"..soext\n"
"    }\n"
"    return traverse_rocks(name, pathes_search)\n"
"end\n"
"\n"
"local function cwd_loader_func(lib)\n"
"    local search_cwd = lib and search_cwd_lib or search_cwd_lua\n"
"    local load_func = lib and load_lib or load_lua\n"
"    return function(name)\n"
"        if not name then\n"
"            return \"empty name of module\"\n"
"        end\n"
"        local file, err = search_cwd(name)\n"
"        if not file then\n"
"            return err\n"
"        end\n"
"        local loaded, err = load_func(file, name)\n"
"        if err == nil then\n"
"            return loaded\n"
"        else\n"
"            return err\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"local function rocks_loader_func(lib)\n"
"    local search_rocks = lib and search_rocks_lib or search_rocks_lua\n"
"    local load_func = lib and load_lib or load_lua\n"
"    return function (name)\n"
"        if not name then\n"
"            return \"empty name of module\"\n"
"        end\n"
"        local file, err = search_rocks(name)\n"
"        if not file then\n"
"            return err\n"
"        end\n"
"        local loaded, err = load_func(file, name)\n"
"        if err == nil then\n"
"            return loaded\n"
"        else\n"
"            return err\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"local function search_path_func(cpath)\n"
"    return function(name)\n"
"        return package.searchpath(name, cpath and package.cpath or package.path)\n"
"    end\n"
"end\n"
"\n"
"local function search(name)\n"
"    if not name then\n"
"        return \"empty name of module\"\n"
"    end\n"
"    local searchers = {\n"
"        search_cwd_lua, search_cwd_lib,\n"
"        search_rocks_lua, search_rocks_lib,\n"
"        search_path_func(false), search_path_func(true)\n"
"    }\n"
"    for _, searcher in ipairs(searchers) do\n"
"        local file = searcher(name)\n"
"        if file ~= nil then\n"
"            return file\n"
"        end\n"
"    end\n"
"    return nil\n"
"end\n"
"\n"
"-- loader_preload 1\n"
"table.insert(package.loaders, 2, cwd_loader_func(false))\n"
"table.insert(package.loaders, 3, cwd_loader_func(true))\n"
"table.insert(package.loaders, 4, rocks_loader_func(false))\n"
"table.insert(package.loaders, 5, rocks_loader_func(true))\n"
"-- package.path   6\n"
"-- package.cpath  7\n"
"-- croot          8\n"
"\n"
"package.search = search\n"
"return {\n"
"    uptime = uptime;\n"
"    pid = pid;\n"
"}\n"
""
;
