const char log_lua[] =
"-- log.lua\n"
"--\n"
"local ffi = require('ffi')\n"
"ffi.cdef[[\n"
"    typedef void (*sayfunc_t)(int level, const char *filename, int line,\n"
"               const char *error, const char *format, ...);\n"
"    void\n"
"    say_set_log_level(int new_level);\n"
"\n"
"    void\n"
"    say_set_log_format(enum say_format format);\n"
"\n"
"\n"
"    extern sayfunc_t _say;\n"
"    extern struct ev_loop;\n"
"    extern struct ev_signal;\n"
"\n"
"    extern void\n"
"    say_logrotate(struct ev_loop *, struct ev_signal *, int);\n"
"\n"
"    enum say_level {\n"
"        S_FATAL,\n"
"        S_SYSERROR,\n"
"        S_ERROR,\n"
"        S_CRIT,\n"
"        S_WARN,\n"
"        S_INFO,\n"
"        S_VERBOSE,\n"
"        S_DEBUG\n"
"    };\n"
"\n"
"    enum say_format {\n"
"        SF_PLAIN,\n"
"        SF_JSON\n"
"    };\n"
"    pid_t log_pid;\n"
"    extern int log_level;\n"
"    extern int log_format;\n"
"]]\n"
"\n"
"local S_WARN = ffi.C.S_WARN\n"
"local S_INFO = ffi.C.S_INFO\n"
"local S_VERBOSE = ffi.C.S_VERBOSE\n"
"local S_DEBUG = ffi.C.S_DEBUG\n"
"local S_ERROR = ffi.C.S_ERROR\n"
"\n"
"local json = require(\"json\").new()\n"
"json.cfg{\n"
"    encode_invalid_numbers = true,\n"
"    encode_load_metatables = true,\n"
"    encode_use_tostring    = true,\n"
"    encode_invalid_as_nil  = true,\n"
"}\n"
"\n"
"local special_fields = {\n"
"    \"file\",\n"
"    \"level\",\n"
"    \"pid\",\n"
"    \"line\",\n"
"    \"cord_name\",\n"
"    \"fiber_name\",\n"
"    \"fiber_id\",\n"
"    \"error_msg\"\n"
"}\n"
"\n"
"local function say(level, fmt, ...)\n"
"    if ffi.C.log_level < level then\n"
"        -- don't waste cycles on debug.getinfo()\n"
"        return\n"
"    end\n"
"    local type_fmt = type(fmt)\n"
"    local format = \"%s\"\n"
"    if select('#', ...) ~= 0 then\n"
"        local stat\n"
"        stat, fmt = pcall(string.format, fmt, ...)\n"
"        if not stat then\n"
"            error(fmt, 3)\n"
"        end\n"
"    elseif type_fmt == 'table' then\n"
"        -- ignore internal keys\n"
"        for _, field in ipairs(special_fields) do\n"
"            fmt[field] = nil\n"
"        end\n"
"        fmt = json.encode(fmt)\n"
"        if ffi.C.log_format == ffi.C.SF_JSON then\n"
"            -- indicate that message is already encoded in JSON\n"
"            format = \"json\"\n"
"        end\n"
"    elseif type_fmt ~= 'string' then\n"
"        fmt = tostring(fmt)\n"
"    end\n"
"\n"
"    local debug = require('debug')\n"
"    local frame = debug.getinfo(3, \"Sl\")\n"
"    local line, file = 0, 'eval'\n"
"    if type(frame) == 'table' then\n"
"        line = frame.currentline or 0\n"
"        file = frame.short_src or frame.src or 'eval'\n"
"    end\n"
"\n"
"    ffi.C._say(level, file, line, nil, format, fmt)\n"
"end\n"
"\n"
"local function say_closure(lvl)\n"
"    return function (fmt, ...)\n"
"        say(lvl, fmt, ...)\n"
"    end\n"
"end\n"
"\n"
"local function log_rotate()\n"
"    ffi.C.say_logrotate(nil, nil, 0)\n"
"end\n"
"\n"
"local function log_level(level)\n"
"    return ffi.C.say_set_log_level(level)\n"
"end\n"
"\n"
"local function log_format(format_name)\n"
"    if format_name == \"json\" then\n"
"        ffi.C.say_set_log_format(ffi.C.SF_JSON)\n"
"    elseif format_name == \"plain\" then\n"
"        ffi.C.say_set_log_format(ffi.C.SF_PLAIN)\n"
"    else\n"
"        error(\"log_format: expected 'json' or 'plain'\")\n"
"    end\n"
"end\n"
"\n"
"local function log_pid()\n"
"    return tonumber(ffi.C.log_pid)\n"
"end\n"
"\n"
"local compat_warning_said = false\n"
"local compat_v16 = {\n"
"    logger_pid = function()\n"
"        if not compat_warning_said then\n"
"            compat_warning_said = true\n"
"            say(S_WARN, 'logger_pid() is deprecated, please use pid() instead')\n"
"        end\n"
"        return log_pid()\n"
"    end;\n"
"}\n"
"\n"
"return setmetatable({\n"
"    warn = say_closure(S_WARN);\n"
"    info = say_closure(S_INFO);\n"
"    verbose = say_closure(S_VERBOSE);\n"
"    debug = say_closure(S_DEBUG);\n"
"    error = say_closure(S_ERROR);\n"
"    rotate = log_rotate;\n"
"    pid = log_pid;\n"
"    level = log_level;\n"
"    log_format = log_format;\n"
"}, {\n"
"    __index = compat_v16;\n"
"})\n"
""
;
