const char uri_lua[] =
"-- uri.lua (internal file)\n"
"\n"
"local ffi = require('ffi')\n"
"\n"
"ffi.cdef[[\n"
"struct uri {\n"
"    const char *scheme;\n"
"    size_t scheme_len;\n"
"    const char *login;\n"
"    size_t login_len;\n"
"    const char *password;\n"
"    size_t password_len;\n"
"    const char *host;\n"
"    size_t host_len;\n"
"    const char *service;\n"
"    size_t service_len;\n"
"    const char *path;\n"
"    size_t path_len;\n"
"    const char *query;\n"
"    size_t query_len;\n"
"    const char *fragment;\n"
"    size_t fragment_len;\n"
"    int host_hint;\n"
"};\n"
"\n"
"int\n"
"uri_parse(struct uri *uri, const char *str);\n"
"\n"
"int\n"
"uri_format(char *str, size_t len, struct uri *uri, bool write_password);\n"
"]]\n"
"\n"
"local builtin = ffi.C;\n"
"\n"
"local uribuf = ffi.new('struct uri')\n"
"\n"
"local function parse(str)\n"
"    if str == nil then\n"
"        error(\"Usage: uri.parse(string)\")\n"
"    end\n"
"    if builtin.uri_parse(uribuf, str) ~= 0 then\n"
"        return nil\n"
"    end\n"
"    local result = {}\n"
"    for _, k in ipairs({ 'scheme', 'login', 'password', 'host', 'service',\n"
"        'path', 'query', 'fragment'}) do\n"
"        if uribuf[k] ~= nil then\n"
"            result[k] = ffi.string(uribuf[k], uribuf[k..'_len'])\n"
"        end\n"
"    end\n"
"    if uribuf.host_hint == 1 then\n"
"        result.ipv4 = result.host\n"
"    elseif uribuf.host_hint == 2 then\n"
"        result.ipv6 = result.host\n"
"    elseif uribuf.host_hint == 3 then\n"
"        result.unix = result.service\n"
"    end\n"
"    return result\n"
"end\n"
"\n"
"local function format(uri, write_password)\n"
"    uribuf.scheme = uri.scheme\n"
"    uribuf.scheme_len = string.len(uri.scheme or '')\n"
"    uribuf.login = uri.login\n"
"    uribuf.login_len = string.len(uri.login or '')\n"
"    uribuf.password = uri.password\n"
"    uribuf.password_len = string.len(uri.password or '')\n"
"    uribuf.host = uri.host\n"
"    uribuf.host_len = string.len(uri.host or '')\n"
"    uribuf.service = uri.service\n"
"    uribuf.service_len = string.len(uri.service or '')\n"
"    uribuf.path = uri.path\n"
"    uribuf.path_len = string.len(uri.path or '')\n"
"    uribuf.query = uri.query\n"
"    uribuf.query_len = string.len(uri.query or '')\n"
"    uribuf.fragment = uri.fragment\n"
"    uribuf.fragment_len = string.len(uri.fragment or '')\n"
"    local str = ffi.new('char[1024]')\n"
"    builtin.uri_format(str, 1024, uribuf, write_password and 1 or 0)\n"
"    return ffi.string(str)\n"
"end\n"
"\n"
"return {\n"
"    parse = parse,\n"
"    format = format,\n"
"};\n"
""
;
